from functools import cache
from pathlib import Path

import requests
from bs4 import BeautifulSoup, Tag

from modules.exceptions import VersionNotFoundError
from modules.updaters.GenericUpdater import GenericUpdater
from modules.utils import parse_hash, sha256_hash_check

DOMAIN = "https://fedoraproject.org"
DOWNLOAD_PAGE_URL = f"{DOMAIN}/spins/[[EDITION]]/download/"
FILE_NAME = "Fedora-[[EDITION]]-Live-x86_64-[[VER]].iso"


class Fedora(GenericUpdater):
    """
    A class representing an updater for Fedora.

    Attributes:
        valid_editions (list[str]): List of valid editions to use
        edition (str): Edition to download
        download_page (requests.Response): The HTTP response containing the download page HTML.
        soup_download_page (BeautifulSoup): The parsed HTML content of the download page.

    Note:
        This class inherits from the abstract base class GenericUpdater.
    """

    def __init__(self, folder_path: Path, edition: str) -> None:
        self.valid_editions = [
            "Budgie",
            "Cinnamon",
            "KDE",
            "LXDE",
            "MATE_Compiz",
            "SoaS",
            "Sway",
            "Xfce",
            "i3",
        ]
        self.edition = edition

        file_path = folder_path / FILE_NAME
        super().__init__(file_path)

        # Make the parameter case insensitive, and find back the correct case using valid_editions
        self.edition = next(
            valid_ed
            for valid_ed in self.valid_editions
            if valid_ed.lower() == self.edition.lower()
        )

        # Weird exception they have
        url_edition = self.edition.lower() if self.edition != "MATE_Compiz" else "mate"

        self.download_page = requests.get(
            DOWNLOAD_PAGE_URL.replace("[[EDITION]]", url_edition)
        )

        if self.download_page.status_code != 200:
            raise ConnectionError(
                f"Failed to fetch the download page from '{self.download_page.url}'"
            )

        self.soup_download_page = BeautifulSoup(
            self.download_page.content, features="html.parser"
        )

    @cache
    def _get_download_link(self) -> str:
        latest_version = self._get_latest_version()
        return f"https://download.fedoraproject.org/pub/fedora/linux/releases/{latest_version[0]}/Spins/x86_64/iso/Fedora-{self.edition}-Live-x86_64-{latest_version[0]}-{latest_version[1]}{'.'+latest_version[2] if len(latest_version)>2 else ''}.iso"

    def check_integrity(self) -> bool:
        latest_version = self._get_latest_version()
        sha256_url = f"https://download.fedoraproject.org/pub/fedora/linux/releases/{latest_version[0]}/Spins/x86_64/iso/Fedora-Spins-{latest_version[0]}-{latest_version[1]}{'.'+latest_version[2] if len(latest_version)>2 else ''}-x86_64-CHECKSUM"

        sha256_sums = requests.get(sha256_url).text

        sha256_sum = parse_hash(sha256_sums, [f"SHA256 (Fedora-{self.edition}"], -1)

        return sha256_hash_check(
            self._get_complete_normalized_file_path(absolute=True),
            sha256_sum,
        )

    @cache
    def _get_latest_version(self) -> list[str]:
        downloads_list: Tag | None = self.soup_download_page.find(
            "div", attrs={"class": "spins-theme"}
        )  # type: ignore
        if not downloads_list:
            raise VersionNotFoundError(
                "We were not able to parse the download categories"
            )

        download_items = downloads_list.find_all(
            "div", attrs={"class": "fp-download-item blue"}
        )
        if not download_items:
            raise VersionNotFoundError(
                "We were not able to parse the list of downloads"
            )

        downloads = next(
            download
            for download in download_items
            if download.find("span", string="Live ISO")
        )
        if not downloads:
            raise VersionNotFoundError(
                "We were not able to parse the Live ISO download item"
            )
        download_a_tag = downloads.find("a", href=True)
        if not download_a_tag:
            raise VersionNotFoundError("We were not able to find the LTS download link")

        return self._str_to_version(
            download_a_tag.get("href")
            .split("x86_64-")[1]
            .split(".iso")[0]
            .replace("-", self.version_splitter)
        )



from functools import cache
from pathlib import Path

import requests
from bs4 import BeautifulSoup, Tag

from modules.exceptions import IntegrityCheckError, VersionNotFoundError
from modules.updaters.GenericUpdater import GenericUpdater
from modules.utils import parse_hash, sha256_hash_check

DOMAIN = "https://clonezilla.org"
FILE_NAME = "clonezilla-live-[[VER]]-amd64.iso"


class Clonezilla(GenericUpdater):
    """
    A class representing an updater for Clonezilla.

    Note:
        This class inherits from the abstract base class GenericUpdater.
    """

    def __init__(self, folder_path: Path) -> None:
        file_path = folder_path / FILE_NAME
        super().__init__(file_path)

    @cache
    def _get_download_link(self) -> str:
        ver = self._version_to_str(self._get_latest_version())
        repo = "https://downloads.sourceforge.net"
        return f"{repo}/clonezilla/clonezilla-live-{Clonezilla._get_clonezilla_version_style(ver)}-amd64.iso"

    def check_integrity(self) -> bool:
        r = requests.get(f"{DOMAIN}/downloads/stable/checksums-contents.php")
        soup = BeautifulSoup(r.content, features="html.parser")
        pre: Tag | None = soup.find("pre")  # type: ignore
        if not pre:
            raise IntegrityCheckError(
                "Unable to extract `<pre>` elements from checksum"
            )

        checksums: list[str] = pre.text.split("###")
        for checksum in checksums:
            if "SHA256" in checksum:
                sha256_sums = checksum
                break
        else:
            raise IntegrityCheckError("Could not find SHA256 sum")

        sha256_hash = parse_hash(sha256_sums, ["amd64.iso"], 0)

        return sha256_hash_check(
            self._get_complete_normalized_file_path(absolute=True), sha256_hash
        )

    @cache
    def _get_latest_version(self) -> list[str]:
        r = requests.get(f"{DOMAIN}/downloads/stable/changelog-contents.php")
        soup = BeautifulSoup(r.content, features="html.parser")
        first_paragraph: Tag | None = soup.find("p")  # type: ignore
        if not first_paragraph:
            raise VersionNotFoundError(
                "Unable to extract `<p>` elements from changelog"
            )
        version = first_paragraph.getText().split()[-1]
        return self._str_to_version(version.replace("-", "."))

    @cache
    @staticmethod
    def _get_clonezilla_version_style(version: str):
        """
        Convert the version string from "x.y.z" to "x.y-z" format, as used by Clonezilla.

        Parameters:
            version (str): The version string in "x.y.z.a" format.

        Returns:
            str: The version string in "x.y.z-a" format.
        """
        return "-".join(version.rsplit(".", 1))



from functools import cache
from pathlib import Path

import requests
from bs4 import BeautifulSoup

from modules.exceptions import VersionNotFoundError
from modules.updaters.GenericUpdater import GenericUpdater
from modules.utils import parse_hash, sha256_hash_check

DOMAIN = "https://mirrors.edge.kernel.org"
DOWNLOAD_PAGE_URL = f"{DOMAIN}/linuxmint/stable/"
FILE_NAME = "linuxmint-[[VER]]-[[EDITION]]-64bit.iso"


class LinuxMint(GenericUpdater):
    """
    A class representing an updater for Linux Mint.

    Attributes:
        valid_editions (list[str]): List of valid editions to use
        edition (str): Edition to download
        download_page (requests.Response): The HTTP response containing the download page HTML.
        soup_download_page (BeautifulSoup): The parsed HTML content of the download page.

    Note:
        This class inherits from the abstract base class GenericUpdater.
    """

    def __init__(self, folder_path: Path, edition: str) -> None:
        self.valid_editions = ["cinnamon", "mate", "xfce"]
        self.edition = edition.lower()

        file_path = folder_path / FILE_NAME
        super().__init__(file_path)

        self.download_page = requests.get(DOWNLOAD_PAGE_URL)

        if self.download_page.status_code != 200:
            raise ConnectionError(
                f"Failed to fetch the download page from '{DOWNLOAD_PAGE_URL}'"
            )

        self.soup_download_page = BeautifulSoup(
            self.download_page.content, features="html.parser"
        )

    @cache
    def _get_download_link(self) -> str:
        latest_version_str = self._version_to_str(self._get_latest_version())
        return f"{DOWNLOAD_PAGE_URL}/{latest_version_str}/{self._get_complete_normalized_file_path(absolute=False)}"

    def check_integrity(self) -> bool:
        latest_version_str = self._version_to_str(self._get_latest_version())

        sha256_url = f"https://mirrors.edge.kernel.org/linuxmint/stable/{latest_version_str}/sha256sum.txt"

        sha256_sums = requests.get(sha256_url).text

        sha256_sum = parse_hash(
            sha256_sums,
            [str(self._get_complete_normalized_file_path(absolute=False))],
            0,
        )

        return sha256_hash_check(
            self._get_complete_normalized_file_path(absolute=True),
            sha256_sum,
        )

    @cache
    def _get_latest_version(self) -> list[str]:
        download_a_tags = self.soup_download_page.find_all("a", href=True)
        if not download_a_tags:
            raise VersionNotFoundError("We were not able to parse the download page")

        local_version = self._get_local_version()
        latest = local_version or []

        for a_tag in download_a_tags:
            href = a_tag.get("href")
            if not href[0].isnumeric():
                continue
            version_number = self._str_to_version(href[:-1])
            if self._compare_version_numbers(latest, version_number):
                latest = version_number

        return latest



import logging
from functools import cache

import requests


@cache
def github_get_latest_version(owner: str, repo: str) -> dict:
    """Gets the latest version of a software via it's GitHub repository

    Args:
        owner (str): Owner of the repository
        repo (str): Name of the repository

    Returns:
        dict: the full release information
    """
    api_url = f"https://api.github.com/repos/{owner}/{repo}"

    logging.debug(f"Fetching latest release from {api_url}")

    release = requests.get(f"{api_url}/releases/latest").json()

    logging.debug(f"GitHub release fetched from {api_url}: {release}")

    return release


def parse_github_release(release: dict) -> dict:
    """Parses a github release into a shorter, easier to read format

    Args:
        release (dict): Release information

    Returns:
        dict: The name of the tag, a dictionary representing {"filename": "url"},
    """
    res = {
        "tag": release["tag_name"],
        "files": {},
        "text": release["body"],
        "source_code": release["zipball_url"],
    }

    for asset in release["assets"]:
        res["files"][asset["name"]] = asset["browser_download_url"]

    logging.debug(f"GitHub release parsed: {res}")

    return res



from functools import cache
from pathlib import Path
from random import shuffle

import requests
from bs4 import BeautifulSoup, Tag

from modules.exceptions import VersionNotFoundError
from modules.updaters.GenericUpdater import GenericUpdater
from modules.utils import parse_hash, sha256_hash_check

DOMAIN = "https://www.ultimatebootcd.com"
DOWNLOAD_PAGE_URL = f"{DOMAIN}/download.html"
MIRRORS = [
    "https://mirror.clientvps.com/ubcd",
    "http://mirror.koddos.net/ubcd",
    "https://mirror.lyrahosting.com/ubcd",
]
FILE_NAME = "ubcd[[VER]].iso"


class UltimateBootCD(GenericUpdater):
    """
    A class representing an updater for Ultimate Boot CD.

    Attributes:
        download_page (requests.Response): The HTTP response containing the download page HTML.
        soup_download_page (BeautifulSoup): The parsed HTML content of the download page.
        mirrors (list[str])
        mirror (str)
        download_table (Tag)

    Note:
        This class inherits from the abstract base class GenericUpdater.
    """

    def __init__(self, folder_path: Path) -> None:
        file_path = folder_path / FILE_NAME
        super().__init__(file_path)

        self.download_page = requests.get(DOWNLOAD_PAGE_URL)

        if self.download_page.status_code != 200:
            raise ConnectionError(
                f"Failed to fetch the download page from '{DOWNLOAD_PAGE_URL}'"
            )

        self.soup_download_page = BeautifulSoup(
            self.download_page.content, features="html.parser"
        )

        self.mirrors = MIRRORS
        shuffle(self.mirrors)

        self.download_table: Tag | None = None
        for mirror in self.mirrors:
            self.mirror_page = requests.get(mirror)

            if self.mirror_page.status_code != 200:
                continue

            self.soup_mirror_page = BeautifulSoup(
                self.mirror_page.content, features="html.parser"
            )

            self.download_table = self.soup_mirror_page.find("table")  # type: ignore
            if self.download_table:
                self.mirror = mirror
                break

        if not self.mirror_page:
            raise ConnectionError(f"Could not connect to any mirrors!")

        if not self.download_table:
            raise LookupError(f"Could not find table of downloads in any mirrors")

    @cache
    def _get_download_link(self) -> str:
        latest_version: list[str] = self._get_latest_version()
        return f"{self.mirror}/ubcd{self._version_to_str(latest_version)}.iso"

    def check_integrity(self) -> bool:
        nowrap_tds: list[Tag] = self.soup_download_page.find_all(
            "td", attrs={"nowrap": "true"}
        )

        tts: list[Tag] = next(td.find_all("tt") for td in nowrap_tds if td.find("tt"))

        sha256_sum: str = next(
            parse_hash(tt.getText(), [], -1) for tt in tts if "SHA-256" in tt.getText()
        )

        return sha256_hash_check(
            self._get_complete_normalized_file_path(absolute=True), sha256_sum
        )

    @cache
    def _get_latest_version(self) -> list[str]:
        download_a_tags = self.download_table.find_all("a", href=True)  # type: ignore
        if not download_a_tags:
            raise VersionNotFoundError("We were not able to parse the download page")

        versions_href = [
            href
            for a_tag in download_a_tags
            if FILE_NAME.split("[[VER]]")[0] in (href := a_tag.get("href"))
            and (href.endswith(".iso"))
        ]

        version = 0
        for version_href in versions_href:
            version_href_number = int("".join(filter(str.isdigit, version_href)))
            if version_href_number > version:
                version = version_href_number

        return self._str_to_version(str(version))



from functools import cache
from pathlib import Path

import requests
from bs4 import BeautifulSoup, Tag

from modules.exceptions import VersionNotFoundError
from modules.updaters.GenericUpdater import GenericUpdater
from modules.utils import parse_hash, sha256_hash_check

DOMAIN = "https://cdimage.debian.org"
DOWNLOAD_PAGE_URL = f"{DOMAIN}/debian-cd/current-live/amd64/iso-hybrid/"
FILE_NAME = "debian-live-[[VER]]-amd64-[[EDITION]].iso"


class Debian(GenericUpdater):
    """
    A class representing an updater for Debian.

    Attributes:
        valid_editions (list[str]): List of valid editions to use
        edition (str): Edition to download
        download_page (requests.Response): The HTTP response containing the download page HTML.
        soup_download_page (BeautifulSoup): The parsed HTML content of the download page.
        soup_index_list (Tag): The index list containing the downloadable files.

    Note:
        This class inherits from the abstract base class GenericUpdater.
    """

    def __init__(self, folder_path: Path, edition: str) -> None:
        self.valid_editions = [
            "cinnamon",
            "gnome",
            "kde",
            "lxde",
            "lxqt",
            "mate",
            "standard",
            "xfce",
        ]

        self.edition = edition.lower()

        file_path = folder_path / FILE_NAME
        super().__init__(file_path)

        # Make the parameter case insensitive, and find back the correct case using valid_editions

        self.download_page = requests.get(DOWNLOAD_PAGE_URL)

        if self.download_page.status_code != 200:
            raise ConnectionError(
                f"Failed to fetch the download page from '{DOWNLOAD_PAGE_URL}'"
            )

        self.soup_download_page = BeautifulSoup(
            self.download_page.content, features="html.parser"
        )

        self.soup_index_list: Tag = self.soup_download_page.find(
            "table", attrs={"id": "indexlist"}
        )  # type: ignore

        if not self.soup_index_list:
            raise ConnectionError(
                "We couldn't find the list of indexes containing the download URLs"
            )

    @cache
    def _get_download_link(self) -> str:
        return f"{DOWNLOAD_PAGE_URL}/{self._get_complete_normalized_file_path(absolute=False)}"

    def check_integrity(self) -> bool:
        sha256_url = f"{DOWNLOAD_PAGE_URL}/SHA256SUMS"

        sha256_sums = requests.get(sha256_url).text

        sha256_sum = parse_hash(
            sha256_sums,
            [str(self._get_complete_normalized_file_path(absolute=False))],
            0,
        )

        return sha256_hash_check(
            self._get_complete_normalized_file_path(absolute=True),
            sha256_sum,
        )

    @cache
    def _get_latest_version(self) -> list[str]:
        download_a_tags = self.soup_index_list.find_all("a", href=True)
        if not download_a_tags:
            raise VersionNotFoundError("We were not able to parse the download page")

        latest = next(
            href
            for a_tag in download_a_tags
            if str(
                self._get_normalized_file_path(
                    absolute=False,
                    version=None,
                    edition=self.edition if self.has_edition() else None,  # type: ignore
                    lang=self.lang if self.has_lang() else None,  # type: ignore
                )
            ).split("[[VER]]")[-1]
            in (href := a_tag.get("href"))
        )

        return self._str_to_version(latest.split("-")[2])



from functools import cache
from pathlib import Path

import requests
from bs4 import BeautifulSoup
from bs4.element import Tag

from modules.exceptions import DownloadLinkNotFoundError, VersionNotFoundError
from modules.updaters.GenericUpdater import GenericUpdater
from modules.utils import sha256_hash_check

DOMAIN = "https://www.hirensbootcd.org/"
DOWNLOAD_PAGE_URL = f"{DOMAIN}/download"
FILE_NAME = "HBCD_PE_[[VER]]_x64.iso"


class HirensBootCDPE(GenericUpdater):
    """
    A class representing an updater for Hiren's Boot CD PE.

    Attributes:
        download_page (requests.Response): The HTTP response containing the download page HTML.
        soup_download_page (BeautifulSoup): The parsed HTML content of the download page.

    Note:
        This class inherits from the abstract base class GenericUpdater.
    """

    def __init__(self, folder_path: Path) -> None:
        file_path = folder_path / FILE_NAME
        super().__init__(file_path)

        self.download_page = requests.get(DOWNLOAD_PAGE_URL)

        if self.download_page.status_code != 200:
            raise ConnectionError(
                f"Failed to fetch the download page from '{DOWNLOAD_PAGE_URL}'"
            )

        self.soup_download_page = BeautifulSoup(
            self.download_page.content, features="html.parser"
        )

    @cache
    def _get_download_link(self) -> str:
        download_tag: Tag | None = self._find_in_table("Filename")

        if not download_tag:
            raise DownloadLinkNotFoundError(
                "Failed to find the `Tag` containing the download link"
            )

        href_attributes = download_tag.find_all(href=True)
        if not href_attributes:
            raise DownloadLinkNotFoundError("No download link found in the `Tag`")

        return href_attributes[0].get("href")

    def check_integrity(self) -> bool:
        """
        Check the integrity of the downloaded file by verifying its SHA-256 hash against the one provided on the website.

        Returns:
            bool: True if the integrity check passes, False otherwise.

        Raises:
            LookupError: If the SHA-256 hash or its container Tag is not found in the download page.
        """
        sha256_tag: Tag | None = self._find_in_table("SHA-256")

        if not sha256_tag:
            raise LookupError("Failed to find the `Tag` containing the SHA-256 value")

        return sha256_hash_check(
            self._get_complete_normalized_file_path(absolute=True), sha256_tag.getText()
        )

    @cache
    def _get_latest_version(self) -> list[str]:
        s: Tag | None = self.soup_download_page.find(
            "div", attrs={"class": "post-content"}
        )  # type: ignore
        if not s:
            raise VersionNotFoundError(
                "Could not find the div containing version information"
            )

        s = s.find("span")  # type: ignore
        if not s:
            raise VersionNotFoundError(
                "Could not find the span containing the version information"
            )

        return self._str_to_version(
            s.getText()
            .split("(v")[1]  # Parse from Hiren’s BootCD PE x64 (v1.0.2) – ISO Content
            .split(")")[0]
        )

    def _find_in_table(self, row_name_contains: str) -> Tag | None:
        """
        Find the HTML Tag containing specific information in the download page table.

        Args:
            row_name_contains (str): A string that identifies the row in the table.

        Returns:
            Tag | None: The HTML Tag containing the desired information, or None if not found.

        Raises:
            LookupError: If the table or the specified row_name_contains is not found in the download page.
        """
        s: Tag | None = self.soup_download_page.find("div", attrs={"class": "table-1"})  # type: ignore

        if not s:
            raise LookupError(
                "Could not find the table containing download information"
            )

        next_is_result = False
        for tr in s.find_all("tr"):
            for td in tr.find_all("td"):
                if next_is_result:
                    return td
                if row_name_contains in td.getText():
                    next_is_result = True

        raise LookupError(f"Failed to find '{row_name_contains}' in the table")



from .ArchLinux import ArchLinux
from .ChromeOS import ChromeOS
from .Clonezilla import Clonezilla
from .Debian import Debian
from .Fedora import Fedora
from .FreeDOS import FreeDOS
from .GenericUpdater import GenericUpdater
from .GPartedLive import GPartedLive
from .HDAT2 import HDAT2
from .HirensBootCDPE import HirensBootCDPE
from .KaliLinux import KaliLinux
from .LinuxMint import LinuxMint
from .Manjaro import Manjaro
from .MemTest86Plus import MemTest86Plus
from .OpenSUSE import OpenSUSE
from .RockyLinux import RockyLinux
from .Tails import Tails
from .Rescuezilla import Rescuezilla
from .ShredOS import ShredOS
from .SuperGrub2 import SuperGrub2
from .SystemRescue import SystemRescue
from .TempleOS import TempleOS
from .Ubuntu import Ubuntu
from .UltimateBootCD import UltimateBootCD
from .TrueNAS import TrueNAS
from .Windows10 import Windows10
from .Windows11 import Windows11

__all__ = [
    "ArchLinux",
    "Debian",
    "ChromeOS",
    "Clonezilla",
    "Fedora",
    "FreeDOS",
    "GenericUpdater",
    "GPartedLive",
    "HDAT2",
    "ShredOS",
    "HirensBootCDPE",
    "KaliLinux",
    "LinuxMint",
    "Manjaro",
    "MemTest86Plus",
    "OpenSUSE",
    "RockyLinux",
    "Tails",
    "Rescuezilla",
    "SuperGrub2",
    "SystemRescue",
    "TempleOS",
    "Ubuntu",
    "UltimateBootCD",
    "TrueNAS",
    "Windows10",
    "Windows11",
]



from functools import cache
from pathlib import Path

from modules.updaters.GenericUpdater import GenericUpdater
from modules.updaters.util_update_checker import (
    github_get_latest_version,
    parse_github_release,
)
from modules.utils import parse_hash, sha1_hash_check

FILE_NAME = " shredos-[[VER]].img "


class ShredOS(GenericUpdater):
    """
    A class representing an updater for ShredOS.

    Attributes:
        valid_editions (list[str]): List of valid editions to use
        edition (str): Edition to download

    Note:
        This class inherits from the abstract base class GenericUpdater.
    """

    def __init__(self, folder_path: Path) -> None:
        file_path = folder_path / FILE_NAME
        super().__init__(file_path)

        release = github_get_latest_version("PartialVolume", "shredos.x86_64")

        self.release_info = parse_github_release(release)

    @cache
    def _get_download_link(self) -> str:
        return next(
            download_link
            for filename, download_link in self.release_info["files"].items()
            if filename.endswith(".img") and "x86-64" in filename
        )

    def check_integrity(self) -> bool:
        sha1_sums = self.release_info["text"]

        sha1_sum = parse_hash(
            sha1_sums,
            [
                "sha1",
                self._version_to_str(self._get_latest_version()),
                "x86-64",
                ".img",
            ],
            1,
        )

        return sha1_hash_check(
            self._get_complete_normalized_file_path(absolute=True),
            sha1_sum,
        )

    @cache
    def _get_latest_version(self) -> list[str]:
        version = self.release_info["tag"]

        start_index = version.find("v")
        end_index = version.find("_")

        version = version[start_index + 1 : end_index]

        return self._str_to_version(version)



from functools import cache
from pathlib import Path

import requests
from bs4 import BeautifulSoup

from modules.exceptions import VersionNotFoundError
from modules.updaters.GenericUpdater import GenericUpdater
from modules.utils import pgp_check

DOMAIN = "https://mirrors.edge.kernel.org"
DOWNLOAD_PAGE_URL = f"{DOMAIN}/tails/stable"
FILE_NAME = "tails-amd64-[[VER]].iso"
PUB_KEY_URL = "https://tails.net/tails-signing.key"


class Tails(GenericUpdater):
    """
    A class representing an updater for Linux Mint.

    Attributes:
        valid_editions (list[str]): List of valid editions to use
        download_page (requests.Response): The HTTP response containing the download page HTML.
        soup_download_page (BeautifulSoup): The parsed HTML content of the download page.

    Note:
        This class inherits from the abstract base class GenericUpdater.
    """

    def __init__(self, folder_path: Path) -> None:
        file_path = folder_path / FILE_NAME
        super().__init__(file_path)

        self.download_page = requests.get(DOWNLOAD_PAGE_URL)

        if self.download_page.status_code != 200:
            raise ConnectionError(
                f"Failed to fetch the download page from '{DOWNLOAD_PAGE_URL}'"
            )

        self.soup_download_page = BeautifulSoup(
            self.download_page.content, features="html.parser"
        )

    @cache
    def _get_download_link(self) -> str:
        latest_version_str = self._version_to_str(self._get_latest_version())
        return f"{DOWNLOAD_PAGE_URL}/tails-amd64-{latest_version_str}/{self._get_complete_normalized_file_path(absolute=False)}"

    def check_integrity(self) -> bool:
        sig_url = f"{self._get_download_link()}.sig"

        sig = requests.get(sig_url).content
        pub_key = requests.get(PUB_KEY_URL).content

        return pgp_check(
            self._get_complete_normalized_file_path(absolute=True), sig, pub_key
        )

    @cache
    def _get_latest_version(self) -> list[str]:
        download_a_tags = self.soup_download_page.find_all("a", href=True)
        if not download_a_tags:
            raise VersionNotFoundError("We were not able to parse the download page")

        local_version = self._get_local_version()
        latest = local_version or []

        for a_tag in download_a_tags:
            href = a_tag.get("href")
            if not "tails-amd64" in href:
                continue
            version = href.split("-")[-1]
            if not version[0].isnumeric():
                continue
            version_number = self._str_to_version(version[:-1])
            if self._compare_version_numbers(latest, version_number):
                latest = version_number

        return latest



from datetime import datetime
from functools import cache
from pathlib import Path

import requests
from bs4 import BeautifulSoup, Tag

from modules.exceptions import VersionNotFoundError
from modules.updaters.GenericUpdater import GenericUpdater
from modules.utils import md5_hash_check, parse_hash

DOMAIN = "https://www.templeos.org"
DOWNLOAD_PAGE_URL = f"{DOMAIN}/Downloads"
FILE_NAME = "TempleOS_[[EDITION]]_[[VER]].ISO"


class TempleOS(GenericUpdater):
    """
    A class representing an updater for TempleOS.

    Attributes:
        valid_editions (list[str]): List of valid editions to use
        edition (str): Edition to download
        download_page (requests.Response): The HTTP response containing the download page HTML.
        soup_download_page (BeautifulSoup): The parsed HTML content of the download page.
        server_file_name (str): The name of the file to download on the server

    Note:
        This class inherits from the abstract base class GenericUpdater.
    """

    def __init__(self, folder_path: Path, edition: str) -> None:
        self.valid_editions = ["Distro", "Lite"]
        self.edition = edition

        file_path = folder_path / FILE_NAME
        super().__init__(file_path)

        # Make the parameter case insensitive, and find back the correct case using valid_editions
        self.edition = next(
            valid_ed
            for valid_ed in self.valid_editions
            if valid_ed.lower() == self.edition.lower()
        )

        self.download_page = requests.get(DOWNLOAD_PAGE_URL)

        if self.download_page.status_code != 200:
            raise ConnectionError(
                f"Failed to fetch the download page from '{DOWNLOAD_PAGE_URL}'"
            )

        self.soup_download_page = BeautifulSoup(
            self.download_page.content, features="html.parser"
        )

        self.server_file_name = (
            f"TempleOS{'Lite' if self.edition == 'Lite' else ''}.ISO"
        )

    @cache
    def _get_download_link(self) -> str:
        return f"{DOWNLOAD_PAGE_URL}/{self.server_file_name}"

    def check_integrity(self) -> bool:
        md5_url = f"{DOWNLOAD_PAGE_URL}/md5sums.txt"

        md5_sums = requests.get(md5_url).text

        md5_sum = parse_hash(md5_sums, [self.server_file_name], 0)

        return md5_hash_check(
            self._get_complete_normalized_file_path(absolute=True),
            md5_sum,
        )

    @cache
    def _get_latest_version(self) -> list[str]:
        file_list_soup: Tag | None = self.soup_download_page.find("pre")  # type: ignore
        if not file_list_soup:
            raise VersionNotFoundError("Could not find download links list")

        page_text = file_list_soup.getText()

        next_line_has_date = False
        date: str | None = None
        for line in page_text.splitlines():
            if self.server_file_name in line:
                next_line_has_date = True
                continue
            if next_line_has_date:
                # Remove last element (size)
                date = " ".join(line.strip().split()[1:-1])
                break
        if not date:
            raise VersionNotFoundError("Could not find date on download page")

        datetime_date = datetime.strptime(date, r"%d-%b-%Y %H:%M")

        return self._str_to_version(str(datetime.timestamp(datetime_date)))



from functools import cache
from pathlib import Path

import requests
from bs4 import BeautifulSoup

from modules.exceptions import VersionNotFoundError
from modules.updaters.GenericUpdater import GenericUpdater
from modules.utils import parse_hash, sha256_hash_check

DOMAIN = "https://releases.ubuntu.com"
DOWNLOAD_PAGE_URL = f"{DOMAIN}"
FILE_NAME = "ubuntu-[[EDITION]]-[[VER]]-desktop-amd64.iso"


class Ubuntu(GenericUpdater):
    """
    A class representing an updater for Ubuntu.

    Attributes:
        valid_editions (list[str]): List of valid editions to use
        edition (str): Edition to download
        download_page (requests.Response): The HTTP response containing the download page HTML.
        soup_download_page (BeautifulSoup): The parsed HTML content of the download page.

    Note:
        This class inherits from the abstract base class GenericUpdater.
    """

    def __init__(self, folder_path: Path, edition: str) -> None:
        self.valid_editions = ["LTS", "Interim"]
        self.edition = edition

        file_path = folder_path / FILE_NAME
        super().__init__(file_path)

        # Make the parameter case insensitive, and find back the correct case using valid_editions
        self.edition = next(
            valid_ed
            for valid_ed in self.valid_editions
            if valid_ed.lower() == self.edition.lower()
        )

        self.download_page = requests.get(DOWNLOAD_PAGE_URL)

        if self.download_page.status_code != 200:
            raise ConnectionError(
                f"Failed to fetch the download page from '{DOWNLOAD_PAGE_URL}'"
            )

        self.soup_download_page = BeautifulSoup(
            self.download_page.content, features="html.parser"
        )

    @cache
    def _get_download_link(self) -> str:
        latest_version_str = self._version_to_str(self._get_latest_version())
        return f"{DOMAIN}/{latest_version_str}/{f'ubuntu-{latest_version_str}-desktop-amd64.iso'}"

    def check_integrity(self) -> bool:
        latest_version_str = self._version_to_str(self._get_latest_version())

        sha256_url = f"{DOWNLOAD_PAGE_URL}/{latest_version_str}/SHA256SUMS"

        sha256_sums = requests.get(sha256_url).text

        sha256_sum = parse_hash(
            sha256_sums, [f"ubuntu-{latest_version_str}-desktop-amd64.iso"], 0
        )

        return sha256_hash_check(
            self._get_complete_normalized_file_path(absolute=True),
            sha256_sum,
        )

    @cache
    def _get_latest_version(self) -> list[str]:
        download_categories = self.soup_download_page.find_all(
            "div", attrs={"class": "col-4"}
        )
        if not download_categories:
            raise VersionNotFoundError(
                "We were not able to parse the download categories"
            )
        downloads = next(
            download_category
            for download_category in download_categories
            if download_category.find("h4", string=f"{self.edition} Releases")
        )
        if not downloads:
            raise VersionNotFoundError(
                f"We were not able to parse the {self.edition} downloads"
            )
        latest = downloads.find("a", href=True)
        if not latest:
            raise VersionNotFoundError(
                f"We were not able to find {self.edition} downloads"
            )
        # This is `x.y`, however to get `x.y.z` we need to go to `/x.y`
        xy_version = latest.getText().split()[1]

        version_page = requests.get(f"{DOWNLOAD_PAGE_URL}/{xy_version}")

        soup_version_page = BeautifulSoup(version_page.content, features="html.parser")

        title = soup_version_page.find("title")

        if not title:
            raise VersionNotFoundError(
                "We were not able to find the title of the version page"
            )

        title_text = title.getText()

        # Extract version from string "Ubuntu x.y.z (Name of Release)"
        return self._str_to_version(title_text.split()[1])



from functools import cache
from pathlib import Path
from urllib.parse import urljoin

import requests
from bs4 import BeautifulSoup
from bs4.element import Tag

from modules.updaters.GenericUpdater import GenericUpdater
from modules.utils import md5_hash_check, parse_hash

DOMAIN = "https://www.hdat2.com"
DOWNLOAD_PAGE_URL = f"{DOMAIN}/download.html"
FILE_NAME = "HDAT2_[[EDITION]]_[[VER]].[[EXT]]"


class HDAT2(GenericUpdater):
    """
    A class representing an updater for HDAT2.

    Attributes:
        valid_editions (list[str]): List of valid editions to use
        edition (str): Edition to download
        download_page (requests.Response): The HTTP response containing the download page HTML.
        soup_download_page (BeautifulSoup): The parsed HTML content of the download page.

    Note:
        This class inherits from the abstract base class GenericUpdater.
    """

    def __init__(self, folder_path: Path, edition: str) -> None:
        self.valid_editions = ["full", "lite", "diskette"]
        self.edition = edition.lower()

        if self.edition == "diskette":
            extension = "IMG"
        else:
            extension = "ISO"

        self.file_name = FILE_NAME.replace("[[EXT]]", extension)

        file_path = folder_path / self.file_name
        super().__init__(file_path)

        self.download_page = requests.get(DOWNLOAD_PAGE_URL)

        if self.download_page.status_code != 200:
            raise ConnectionError(
                f"Failed to fetch the download page from '{DOWNLOAD_PAGE_URL}'"
            )

        self.soup_download_page = BeautifulSoup(
            self.download_page.content, features="html.parser"
        )

    @cache
    def _get_download_link(self) -> str:
        version_str = self._version_to_str(self._get_latest_version())
        match self.edition:
            case "full":
                soup = self._find_in_table([version_str], ["LITE", "IMG", "EXE"])
            case "lite":
                soup = self._find_in_table([version_str, "LITE"], ["IMG", "EXE"])
            case "diskette":
                soup = self._find_in_table([version_str, "HDAT2IMG"], ["ISO", "EXE"])
            case _:
                raise NotImplementedError(
                    f"Edition {self.edition} is not implemented yet."
                )

        a_tag = soup.find("a", href=True)

        if not a_tag:
            raise LookupError("Could not find HTML tag containing download link")

        return urljoin(DOMAIN, a_tag["href"])  # type: ignore

    def check_integrity(self) -> bool:
        version_str = self._version_to_str(self._get_latest_version())
        match self.edition:
            case "full":
                soup = self._find_in_table([version_str], ["LITE"])
            case "lite":
                soup = self._find_in_table([version_str, "LITE"])
            case "diskette":
                soup = self._find_in_table([version_str, "HDAT2IMG"])
            case _:
                raise NotImplementedError(
                    f"Edition {self.edition} is not implemented yet."
                )
        tag_with_hash = soup.find(lambda tag: "MD5=" in tag.text)
        if not tag_with_hash:
            raise LookupError("Could not find HTML tag containing MD5 hash")

        md5_sum = parse_hash(tag_with_hash.text, ["MD5=", version_str], -1).replace(
            "MD5=", ""
        )

        return md5_hash_check(
            self._get_complete_normalized_file_path(absolute=True),
            md5_sum,
        )

    @cache
    def _get_latest_version(self) -> list[str]:
        version_tag = self.soup_download_page.find("font", {"color": "blue"})

        if not version_tag:
            raise LookupError(
                "Could not find the HTML tag containing the version number"
            )

        version_text = version_tag.get_text(strip=True)

        return self._str_to_version(
            version_text.split()[2]  # Get 'x.y' from 'Latest version x.y date'
        )

    def _find_in_table(
        self,
        row_name_contains: list[str],
        row_name_doesnt_contain: list[str] | None = None,
    ) -> Tag:
        """
        Find the HTML Tag containing specific information in the download page table.

        Args:
            row_name_contains (list[str]): Strings that identify the row in the table.
            row_name_doesnt_contain (list[str]): Strings that shouldn't be in the row

        Returns:
            Tag: The HTML Tag containing the desired information, or None if not found.

        Raises:
            LookupError: If the table or the specified row_name_contains is not found in the download page.
        """
        if not row_name_doesnt_contain:
            row_name_doesnt_contain = []
        s: Tag | None = self.soup_download_page.find("table", attrs={"bgcolor": "#B3B3B3"})  # type: ignore

        if not s:
            raise LookupError(
                "Could not find the table containing download information"
            )

        for tr in s.find_all("tr"):
            text = tr.getText()
            if any(string in text for string in row_name_doesnt_contain):
                continue
            if all(string in text for string in row_name_contains):
                return tr

        raise LookupError(f"Failed to find value in the table")



import glob
import logging
import re
from abc import ABC, abstractmethod
from pathlib import Path

from modules.exceptions import IntegrityCheckError
from modules.utils import download_file


class GenericUpdater(ABC):
    """
    Abstract base class for a generic updater that manages software updates.

    Attributes:
        file_path (Path): The path to the file that needs to be updated.
    """

    def __init__(self, file_path: Path, *args, **kwargs) -> None:
        """
        Initialize the GenericUpdater instance.

        Args:
            file_path (Path): The path to the file that needs to be updated.
        """
        self.file_path = file_path.resolve()
        self.folder_path = file_path.parent.resolve()

        self.version_splitter = "."

        if self.has_edition():
            logging.debug(
                f"[GenericUpdater.__init__] {self.__class__.__name__} has edition support"
            )
            if self.edition.lower() not in (  # type: ignore
                valid_edition.lower() for valid_edition in self.valid_editions  # type: ignore
            ):
                raise ValueError(
                    f"Invalid edition. The available editions are: {', '.join(self.valid_editions)}."  # type: ignore
                )

        if self.has_lang():
            logging.debug(
                f"[GenericUpdater.__init__] {self.__class__.__name__} has language support"
            )
            if self.lang.lower() not in (  # type: ignore
                valid_lang.lower() for valid_lang in self.valid_langs  # type: ignore
            ):
                raise ValueError(
                    f"Invalid language. The available languages are: {', '.join(self.valid_langs)}."  # type: ignore
                )

        self.folder_path.mkdir(parents=True, exist_ok=True)

    @abstractmethod
    def _get_download_link(self) -> str:
        """
        (Protected) Get the download link for the latest version of the software.

        Returns:
            str: The download link for the latest version of the software.

        Raises:
            DownloadLinkNotFoundError: If the download link is not found.
        """
        pass

    @abstractmethod
    def check_integrity(self) -> bool:
        """
        Check the integrity of the downloaded software.

        Returns:
            bool: True if the downloaded software is valid, otherwise False.
        """
        pass

    def check_for_updates(self) -> bool:
        """
        Check if there are updates available for the software.

        Returns:
            bool: True if updates are available, False if the local version is up to date.
        """
        if not (local_version := self._get_local_version()):
            logging.debug(
                f"[GenericUpdater.check_for_updates] No local version found for {self.__class__.__name__}"
            )
            return True

        is_update_available = self._compare_version_numbers(
            local_version, self._get_latest_version()
        )
        logging.debug(
            f"[GenericUpdater.check_for_updates] {self._version_to_str(local_version)} > {self._version_to_str(self._get_latest_version())}? {is_update_available}"
        )
        return is_update_available

    def install_latest_version(self) -> None:
        """
        Install the latest version of the software.

        Raises:
            IntegrityCheckError: If the integrity check of the downloaded file fails.
        """
        download_link = self._get_download_link()

        # Determine the old and new file paths
        old_file = self._get_local_file()
        new_file = self._get_complete_normalized_file_path(absolute=True)

        if not self.has_version():
            # If the file is being replaced, back it up
            if old_file:
                logging.debug(
                    f"[GenericUpdater.install_latest_version] Renaming old file: {old_file}"
                )
                old_file.with_suffix(".old").replace(old_file)

        download_file(download_link, new_file)

        # Check the integrity of the downloaded file
        try:
            integrity_check = self.check_integrity()
        except Exception as e:
            # If integrity check failed, restore the old file or remove the new file
            if self.has_version() or not old_file:
                new_file.unlink()
            else:
                old_file.replace(new_file)
            raise IntegrityCheckError(
                "Integrity check failed: An error occurred"
            ) from e

        if not integrity_check:
            # If integrity check failed, restore the old file or remove the new file
            if self.has_version() or not old_file:
                new_file.unlink()
            else:
                old_file.replace(new_file)
            raise IntegrityCheckError("Integrity check failed: Hashes do not match")

        # If the installation was successful and we had a previous version installed, remove it
        if old_file:
            logging.debug(
                f"[GenericUpdater.install_latest_version] Removing old file: {old_file}"
            )
            old_file.unlink()

    def has_version(self) -> bool:
        """
        Check if the updater supports different versions.

        Returns:
            bool: True if different versions are supported, False otherwise.
        """
        return "[[VER]]" in str(self.file_path)

    def has_edition(self) -> bool:
        """
        Check if the updater supports different editions.

        Returns:
            bool: True if different editions are supported, False otherwise.
        """
        return (
            hasattr(self, "edition")
            and hasattr(self, "valid_editions")
            and "[[EDITION]]" in str(self.file_path)
        )

    def has_lang(self) -> bool:
        """
        Check if the updater supports different languages.

        Returns:
            bool: True if different languages are supported, False otherwise.
        """
        return (
            hasattr(self, "lang")
            and hasattr(self, "valid_langs")
            and "[[LANG]]" in str(self.file_path)
        )

    def _get_local_file(self) -> Path | None:
        """
        Get the path of the locally stored file that matches the filename pattern.

        Returns:
            str | None: The path of the locally stored file if found, None if no file exists.
        """
        file_path = self._get_normalized_file_path(
            absolute=True,
            version=None,
            edition=self.edition if self.has_edition() else None,  # type: ignore
            lang=self.lang if self.has_lang() else None,  # type: ignore
        )

        local_files = glob.glob(str(file_path).replace("[[VER]]", "*"))

        if local_files:
            return Path(local_files[0])
        logging.debug(
            f"[GenericUpdater._get_local_file] No local file found for {self.__class__.__name__}"
        )
        return None

    def _get_local_version(self) -> list[str] | None:
        """
        Get the version of the locally stored file by extracting the version number from the filename.

        Returns:
            list[str] | None: A list of integers representing the version number if found,
                            None if the version cannot be determined or no local file exists.
        """
        local_version: list[str] | None = None

        local_file = self._get_local_file()

        if not local_file or not self.has_version():
            logging.debug(
                f"[GenericUpdater._get_local_version] No local version found for {self.__class__.__name__}"
            )
            return None

        normalized_path_without_ext = Path(
            self._get_normalized_file_path(
                absolute=True,
                version=None,
                edition=self.edition if self.has_edition() else None,  # type: ignore
                lang=self.lang if self.has_lang() else None,  # type: ignore
            )
        ).with_suffix("")

        version_regex: str = r"(.+)".join(
            re.escape(part)
            for part in str(normalized_path_without_ext).split("[[VER]]")
        )
        local_version_regex = re.search(version_regex, str(local_file))

        if local_version_regex:
            local_version = self._str_to_version(local_version_regex.group(1))

        if not local_version:
            logging.debug(
                f"[GenericUpdater._get_local_version] No local version found for {self.__class__.__name__}"
            )

        return local_version

    def _get_latest_version(self) -> list[str]:
        """
        Get the latest version of the software from the download page.

        Returns:
            list[str]: A list of integers representing the latest version number.

        Raises:
            VersionNotFoundError: If the latest version cannot be found on the download page.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} has not been implemented yet."
        )

    def _get_normalized_file_path(
        self,
        absolute: bool,
        version: list[str] | None = None,
        edition: str | None = None,
        lang: str | None = None,
    ) -> Path:
        """
        Get the normalized file path with customizable version, edition, and language.

        Args:
            absolute (bool): If True, return the absolute file path. Otherwise, return the relative file path.
            version (list[str], optional): The version as a list of version components.
                                    If provided, it replaces '[[VER]]' in the file name.
                                    Defaults to None.
            edition (str, optional): The edition of the file. If provided, it replaces '[[EDITION]]' in the file name.
                                    Defaults to None.
            lang (str, optional): The language of the file. If provided, it replaces '[[LANG]]' in the file name.
                                    Defaults to None.

        Returns:
            Path: The normalized file path.

        Note:
            This method replaces placeholders such as '[[VER]]', '[[EDITION]]', and '[[LANG]]' in the file name
            with the specified version, edition, and language respectively. It also removes all spaces from the file name.
        """
        file_name: str = self.file_path.name

        # Replace placeholders with the specified version, edition, and language
        if version is not None and "[[VER]]" in file_name:
            file_name = file_name.replace("[[VER]]", self._version_to_str(version))

        if edition is not None and "[[EDITION]]" in file_name:
            file_name = file_name.replace("[[EDITION]]", edition)

        if lang is not None and "[[LANG]]" in file_name:
            file_name = file_name.replace("[[LANG]]", lang)

        # Remove all spaces from the file name
        file_name = "".join(file_name.split())

        # Return the absolute or relative file path based on the 'absolute' parameter
        return self.folder_path / file_name if absolute else Path(file_name)

    def _get_complete_normalized_file_path(
        self, absolute: bool, latest: bool = True
    ) -> Path:
        """
        Get the complete normalized file path with customizable version, edition, and language.

        Args:
            absolute (bool): If True, return the absolute file path. Otherwise, return the relative file path.
            latest (bool, optional): If True, use the latest version, edition, and language to construct the file path.
                                    If False, use the local version, edition, and language.
                                    Defaults to True.

        Returns:
            Path: The normalized file path.

        Note:
            This method replaces placeholders such as '[[VER]]', '[[EDITION]]', and '[[LANG]]' in the file name
            with the specified version, edition, and language respectively. It also removes all spaces from the file name.
        """
        return self._get_normalized_file_path(
            absolute=absolute,
            version=self._get_latest_version() if latest else self._get_local_version(),
            edition=self.edition if self.has_edition() else None,  # type: ignore
            lang=self.lang if self.has_lang() else None,  # type: ignore
        )

    def _version_to_str(self, version: list[str]):
        """
        Convert a list of version components to a version string.

        Args:
            version (list[str]): The version as a list of version components.

        Returns:
            str: The version as a string with components joined by the version splitter.
        """
        return self.version_splitter.join(str(i) for i in version)

    def _str_to_version(self, version_str: str):
        """
        Convert a version string to a list of version components.

        Args:
            version_str (str): The version as a string.

        Returns:
            list[str]: The version as a list of version components.
        """
        return [
            version_number.strip()
            for version_number in version_str.split(self.version_splitter)
        ]

    @staticmethod
    def _compare_version_numbers(
        old_version: list[str], new_version: list[str]
    ) -> bool:
        """
        Compare version numbers to check if a new version is available.

        Args:
            old_version (list[str]): The old version as a list of version components.
            new_version (list[str]): The new version as a list of version components.

        Returns:
            bool: True if the new version is greater than the old version, False otherwise.
        """
        for i in range(len(new_version)):
            try:
                if int(new_version[i]) > int(old_version[i]):
                    return True
            except ValueError:
                if int(new_version[i], 32) > int(old_version[i], 32):
                    return True
            except IndexError:
                return True
        return False



import hashlib
import logging
import re
import shutil
import tomllib
import traceback
import uuid
from pathlib import Path

import requests
from bs4 import BeautifulSoup, Tag
from pgpy import PGPKey, PGPSignature
from tqdm import tqdm

READ_CHUNK_SIZE = 524288


def logging_critical_exception(msg, *args, **kwargs):
    """
    Log a critical exception with traceback information.

    Args:
        msg (str): The error message to be logged.
        *args: Variable length argument list to be passed to the logging.critical method.
        **kwargs: Keyword arguments to be passed to the logging.critical method.
    """
    logging.critical(f"{msg}\n{traceback.format_exc()}", *args, **kwargs)


def parse_config(toml_file: Path) -> dict | None:
    """Parse a TOML configuration file and return a dictionary representation.

    Args:
        toml_file (Path): The path to the TOML configuration file.

    Returns:
        dict | None: The parsed configuration as a dictionary, or None if there was an error during parsing.
    """
    with open(toml_file, "rb") as f:
        toml_dict = tomllib.load(f)
    return parse_config_from_dict(toml_dict)


def parse_config_from_dict(input_dict: dict):
    """Recursively parse the nested config dictionary and return a new dictionary where the keys are the directory, unless they are a module's name.

    Args:
        input_dict (dict): The input dictionary to be parsed.

    Returns:
        dict: The parsed dictionary with modified keys.
    """
    new_dict = {}
    for key, value in input_dict.items():
        if isinstance(value, dict):
            if "enabled" in value and not value["enabled"]:
                logging.debug(f"Skipping disabled module {key}")
                del value
                continue
            if "directory" in value:
                logging.debug(f"Found directory {value['directory']}")
                new_key = value["directory"]
                del value["directory"]
            else:
                logging.debug(f"Found module {key}")
                new_key = key
            new_dict[new_key] = parse_config_from_dict(value)
        elif key == "enabled":
            continue
        else:
            logging.debug(f"Found key {key}")
            new_dict[key] = value
    return new_dict


def md5_hash_check(file: Path, hash: str) -> bool:
    """
    Calculate the MD5 hash of a given file and compare it with a provided hash value.

    Args:
        file (Path): The path to the file for which the hash is to be calculated.
        hash (str): The MD5 hash value to compare against the calculated hash.

    Returns:
        bool: True if the calculated MD5 hash matches the provided hash; otherwise, False.
    """
    with open(file, "rb") as f:
        file_hash = hashlib.md5()
        while chunk := f.read(READ_CHUNK_SIZE):
            file_hash.update(chunk)
    result = hash.lower() == file_hash.hexdigest()

    logging.debug(
        f"[md5_hash_check] {file.resolve()}: `{hash.lower()}` is {'' if result else 'not'} equal to file hash `{file_hash.hexdigest()}`"
    )
    return result


def sha1_hash_check(file: Path, hash: str) -> bool:
    """
    Calculate the SHA-1 hash of a given file and compare it with a provided hash value.

    Args:
        file (Path): The path to the file for which the hash is to be calculated.
        hash (str): The SHA-1 hash value to compare against the calculated hash.

    Returns:
        bool: True if the calculated SHA-1 hash matches the provided hash; otherwise, False.
    """
    with open(file, "rb") as f:
        file_hash = hashlib.sha1()
        while chunk := f.read(READ_CHUNK_SIZE):
            file_hash.update(chunk)
    result = hash.lower() == file_hash.hexdigest()

    logging.debug(
        f"[sha1_hash_check] {file.resolve()}: `{hash.lower()}` is {'' if result else 'not'} equal to file hash `{file_hash.hexdigest()}`"
    )
    return result


def sha256_hash_check(file: Path, hash: str) -> bool:
    """
    Calculate the SHA-256 hash of a given file and compare it with a provided hash value.

    Args:
        file (str): The path to the file for which the hash is to be calculated.
        hash (str): The SHA-256 hash value to compare against the calculated hash.

    Returns:
        bool: True if the calculated SHA-256 hash matches the provided hash; otherwise, False.
    """
    with open(file, "rb") as f:
        file_hash = hashlib.sha256()
        while chunk := f.read(READ_CHUNK_SIZE):
            file_hash.update(chunk)
    result = hash.lower() == file_hash.hexdigest()

    logging.debug(
        f"[sha256_hash_check] {file.resolve()}: `{hash.lower()}` is {'' if result else 'not'} equal to file hash `{file_hash.hexdigest()}`"
    )
    return result


def sha512_hash_check(file: Path, hash: str) -> bool:
    """
    Calculate the SHA-512 hash of a given file and compare it with a provided hash value.

    Args:
        file (Path): The path to the file for which the hash is to be calculated.
        hash (str): The SHA-512 hash value to compare against the calculated hash.

    Returns:
        bool: True if the calculated SHA-512 hash matches the provided hash; otherwise, False.
    """
    with open(file, "rb") as f:
        file_hash = hashlib.sha512()
        while chunk := f.read(READ_CHUNK_SIZE):
            file_hash.update(chunk)
    result = hash.lower() == file_hash.hexdigest()

    logging.debug(
        f"[sha512_hash_check] {file.resolve()}: `{hash.lower()}` is {'' if result else 'not'} equal to file hash `{file_hash.hexdigest()}`"
    )
    return result


def pgp_check(file_path: Path, signature: str | bytes, public_key: str | bytes) -> bool:
    """Verifies the signature of a file against a publick ey

    Args:
        file_path (Path): Path to the file to check
        signature (str | bytes): Signature
        public_key (str | bytes): Public Key

    Raises:
        ValueError: If the supplied public key is invalid
        ValueError: If the supplied signature is invalid

    Returns:
        bool: Weither the check was successful or not
    """
    pub_key = PGPKey.from_blob(public_key)
    sig = PGPSignature.from_blob(signature)

    if not pub_key:
        raise ValueError(f"Invalid pub_key: {public_key}")
    elif not sig:
        raise ValueError(f"Invalid signature: {signature}")

    # For some reason, from_blob can return either a tuple with either [ThingIwant, Literally Nothing] or directly ThingIWant
    if isinstance(pub_key, tuple):
        pub_key = pub_key[0]
    if isinstance(sig, tuple):
        sig = sig[0]

    with open(file_path, "rb") as f:
        file_content = f.read()

    result = bool(pub_key.verify(file_content, sig))

    logging.debug(
        f"[pgp_check] {file_path.resolve()}: Signature is {'' if result else 'not'} valid"
    )

    return result


def parse_hash(
    hashes: str, match_strings_in_line: list[str], hash_position_in_line: int
):
    """Parse a list of hashes and extract a specific hash based on matching strings.

    Args:
        hashes (str): A string containing a list of hashes.
        match_strings_in_line (list[str]): List of strings that must be present in the line to consider it.
        hash_position_in_line (int): The position of the desired hash in each line.

    Returns:
        The extracted hash value.
    """
    logging.debug(
        f"[parse_hash] Parsing hashes with match strings `{match_strings_in_line}` and hash position {hash_position_in_line} in those hashes:\n{hashes}"
    )
    hash = next(
        line.split()[hash_position_in_line]
        for line in hashes.strip().splitlines()
        if all(match in line for match in match_strings_in_line)
    )
    logging.debug(f"[parse_hash] Extracted hash: `{hash}`")
    return hash


def download_file(url: str, local_file: Path, progress_bar: bool = True) -> None:
    """
    Download a file from a given URL and save it to the local file system.

    Args:
        url (str): The URL of the file to download.
        local_file (Path): The path where the downloaded file will be saved on the local file system.
        progress_bar (bool): Whether to show a progress bar during the download (default: True).

    Returns:
        None
    """
    part_file = local_file.with_suffix(".part")
    logging.debug(f"[download_file] Downloading {url} to {part_file.resolve()}")

    try:
        with requests.get(url, stream=True) as r:
            total_size = int(r.headers.get("content-length", 0))  # Sizes in bytes

            with open(part_file, "wb") as f:
                if progress_bar:
                    with tqdm(
                        total=total_size,
                        unit="B",
                        desc=part_file.name,
                    ) as pbar:
                        for chunk in r.iter_content(chunk_size=1024):
                            if chunk:
                                f.write(chunk)
                                pbar.update(len(chunk))
                else:
                    shutil.copyfileobj(r.raw, f)
    except requests.exceptions.RequestException:
        logging.exception(f"Failed to download {url} to {part_file.resolve()}")
        if part_file.exists():
            part_file.unlink()
        raise
    except KeyboardInterrupt:
        logging.info(f"Download of {url} to {part_file.resolve()} was cancelled")
        if part_file.exists():
            part_file.unlink()
        raise

    part_file.rename(local_file)


def windows_consumer_download(
    windows_version: str = "11", lang="English International"
) -> tuple[str, str]:
    """
    Obtain a Windows ISO download URL for a specific Windows version and language.

    Args:
        windows_version (str): The desired Windows version. Valid options are '11', '10', or '8'.
                            Default is '11'.
        lang (str): The desired language for the Windows ISO. Default is 'English International'.
            See https://www.microsoft.com/en-us/software-download/windows11 for a list of available languages

    Returns:
        tuple[str, str]: A tuple containing the download link and the SHA256 hash of the downloaded file.
    """
    # Thanks https://github.com/pbatard/Fido/ for already having done the reverse-engineering for me <3
    match windows_version:
        case "11":
            url_segment = f"windows{windows_version}"
        case "10":
            url_segment = f"windows{windows_version}ISO"
        case "8":
            url_segment = f"windows{windows_version}ISO"
        case _:
            raise NotImplementedError(
                "The valid windows versions are '11', '10', or '8'."
            )

    url = f"https://www.microsoft.com/en-us/software-download/{url_segment}"

    headers = {
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:115.0) Gecko/20100101 Firefox/115.0",
        "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,*/*;q=0.8",
        "referer": "localhost",
    }
    session_id = uuid.uuid4()

    r = requests.get(url, headers=headers)

    matches = re.search(r'<option value="([0-9]+)">Windows', r.text)
    if not matches or not matches.groups():
        raise LookupError("Could not find product edition id")

    product_edition_id = matches.group(1)

    logging.debug(
        f"[windows_consumer_download] Product edition id: `{product_edition_id}`"
    )

    # Permit Session ID
    r = requests.get(
        f"https://vlscppe.microsoft.com/tags?org_id=y6jn8c31&session_id={session_id}",
        headers=headers,
    )

    language_skuID_table_html_url = f"https://www.microsoft.com/en-US/api/controls/contentinclude/html?pageId=a8f8f489-4c7f-463a-9ca6-5cff94d8d041&host=www.microsoft.com&segments=software-download,{url_segment}&query=&action=getskuinformationbyproductedition&sessionId={session_id}&productEditionId={product_edition_id}&sdVersion=2"

    language_skuID_table_html = requests.get(
        language_skuID_table_html_url, headers=headers
    )

    sku_id = "".join(
        filter(
            str.isdigit,
            next(
                table_line
                for table_line in language_skuID_table_html.text.splitlines()
                if lang in table_line
            ),
        )
    )

    logging.debug(f"[windows_consumer_download] Found SKU ID {sku_id} for {lang}")

    # Get ISO download link page
    iso_download_link_page = f"https://www.microsoft.com/en-US/api/controls/contentinclude/html?pageId=6e2a1789-ef16-4f27-a296-74ef7ef5d96b&host=www.microsoft.com&segments=software-download,{url_segment}&query=&action=GetProductDownloadLinksBySku&sessionId={session_id}&skuId={sku_id}&language={lang}&sdVersion=2"
    r = requests.get(iso_download_link_page, headers=headers)

    s = BeautifulSoup(r.content, features="html.parser")

    hash_table: Tag | None = s.find("table", "table-bordered")  # type: ignore
    if not hash_table:
        raise LookupError(
            "Could not find table containing SHA256 hashes. (Are you being rate limited?)"
        )

    next_is_result = False
    result: Tag | None = None
    for tr in hash_table.find_all("tr"):
        if result:
            break
        for td in tr.find_all("td"):
            if next_is_result:
                result = td
                break
            if lang in td.getText():
                next_is_result = True
    if not result:
        raise LookupError(
            "Could not find row containing SHA256 hashes. (Are you being rate limited?)"
        )

    hash = result.getText()

    download_a_tag: Tag | None = s.find(
        "a", href=True, attrs={"class": "button button-long button-flat button-purple"}
    )  # type: ignore
    if not download_a_tag:
        raise LookupError(
            "Could not find tag containing the download link. (Are you being rate limited?)"
        )
    download_link: str = download_a_tag.get("href")  # type: ignore

    return download_link, hash



import zipfile
from functools import cache
from pathlib import Path

import requests

from modules.exceptions import IntegrityCheckError
from modules.updaters.GenericUpdater import GenericUpdater
from modules.utils import download_file, sha1_hash_check

DOMAIN = "https://dl.google.com"
FILE_NAME = "chromeos_[[VER]]_[[EDITION]].img"


class ChromeOS(GenericUpdater):
    """
    A class representing an updater for ChromeOS.

    Attributes:
        valid_editions (list[str]): List of valid editions to use
        edition (str): Edition to download
        chromium_releases_info (list[dict]): List of release info for each edition
        cur_edition_info: Release info for the selected edition

    Note:
        This class inherits from the abstract base class GenericUpdater.
    """

    def __init__(self, folder_path: Path, edition: str) -> None:
        self.valid_editions = ["ltc", "ltr", "stable"]
        self.edition = edition.lower()

        file_path = Path(folder_path) / FILE_NAME
        super().__init__(file_path)

        self.chromium_releases_info: list[dict] = requests.get(
            f"{DOMAIN}/dl/edgedl/chromeos/recovery/cloudready_recovery2.json"
        ).json()

        self.cur_edition_info: dict = next(
            d
            for d in self.chromium_releases_info
            if d["channel"].lower() == self.edition
        )

    @cache
    def _get_download_link(self) -> str:
        return self.cur_edition_info["url"]

    def check_integrity(self) -> bool:
        sha1_sum = self.cur_edition_info["sha1"]

        return sha1_hash_check(
            self._get_complete_normalized_file_path(absolute=True).with_suffix(".zip"),
            sha1_sum,
        )

    def install_latest_version(self) -> None:
        """
        Download and install the latest version of the software.

        Raises:
            IntegrityCheckError: If the integrity check of the downloaded file fails.
        """
        download_link = self._get_download_link()

        new_file = self._get_complete_normalized_file_path(absolute=True)

        archive_path = Path(new_file).with_suffix(".zip")

        local_file = self._get_local_file()

        download_file(download_link, archive_path)

        try:
            integrity_check = self.check_integrity()
        except Exception as e:
            raise IntegrityCheckError(
                "Integrity check failed: An error occurred"
            ) from e

        if not integrity_check:
            archive_path.unlink()
            raise IntegrityCheckError("Integrity check failed: Hashes do not match")

        with zipfile.ZipFile(archive_path) as z:
            file_list = z.namelist()

            file_ext = "bin"
            to_extract = next(
                file for file in file_list if file.lower().endswith(file_ext)
            )

            extracted_file = Path(z.extract(to_extract, path=new_file.parent))
        try:
            extracted_file.rename(new_file)
        except FileExistsError:
            # On Windows, files are not overwritten by default, so we need to remove the old file first
            new_file.unlink()
            extracted_file.rename(new_file)

        archive_path.unlink()
        if local_file:
            local_file.unlink()

    @cache
    def _get_latest_version(self) -> list[str]:
        return self._str_to_version(self.cur_edition_info["version"])



from functools import cache
from pathlib import Path

import requests

from modules.exceptions import IntegrityCheckError
from modules.updaters.GenericUpdater import GenericUpdater
from modules.utils import parse_hash, sha256_hash_check

DOMAIN = "https://gparted.org"
FILE_NAME = "gparted-live-[[VER]]-amd64.iso"


class GPartedLive(GenericUpdater):
    """
    A class representing an updater for GParted Live.

    Note:
        This class inherits from the abstract base class GenericUpdater.
    """

    def __init__(self, folder_path: Path) -> None:
        file_path = folder_path / FILE_NAME
        super().__init__(file_path)

        self.checksum_file: str = requests.get(
            "https://gparted.org/gparted-live/stable/CHECKSUMS.TXT"
        ).text.strip()

    @cache
    def _get_download_link(self) -> str:
        ver = self._version_to_str(self._get_latest_version())
        return f"https://downloads.sourceforge.net/gparted/gparted-live-{GPartedLive._get_gparted_version_style(ver)}-amd64.iso"

    def check_integrity(self) -> bool:
        checksums: list[str] = self.checksum_file.split("###")
        for checksum in checksums:
            if "SHA256" in checksum:
                sha256_sums = checksum
                break
        else:
            raise IntegrityCheckError("Could not find SHA256 sum")

        sha256_hash = parse_hash(sha256_sums, ["amd64.iso"], 0)

        return sha256_hash_check(
            self._get_complete_normalized_file_path(absolute=True), sha256_hash
        )

    @cache
    def _get_latest_version(self) -> list[str]:
        # Get last line of the checksum file
        version = self.checksum_file.splitlines()[-1]
        # Get last "word" of the line (Which is the file name)
        version = version.split()[-1]
        # Only keep the version numbers and join them with a dot in between each of them
        version = ".".join(version.split("-")[2:4])

        return self._str_to_version(version)

    @cache
    @staticmethod
    def _get_gparted_version_style(version: str):
        """
        Convert the version string from "x.y.z.a" to "x.y.z-a" format, as used by GParted Live.

        Parameters:
            version (str): The version string in "x.y.z.a" format.

        Returns:
            str: The version string in "x.y.z-a" format.
        """
        return "-".join(version.rsplit(".", 1))



from functools import cache
from pathlib import Path

import requests

from modules.updaters.GenericUpdater import GenericUpdater
from modules.utils import parse_hash, sha256_hash_check

DOMAIN = "https://download.opensuse.org"
DOWNLOAD_PAGE_URL = f"{DOMAIN}/download/distribution/[[EDITION]]"
FILE_NAME = "openSUSE-[[EDITION]]-[[VER]]-DVD-x86_64-Current.iso"


class OpenSUSE(GenericUpdater):
    """
    A class representing an updater for OpenSUSE.

    Attributes:
        valid_editions (list[str]): List of valid editions to use
        edition (str): Edition to download

    Note:
        This class inherits from the abstract base class GenericUpdater.
    """

    def __init__(self, folder_path: Path, edition: str) -> None:
        self.valid_editions = ["leap", "leap-micro", "jump"]
        self.edition = edition.lower()

        self.download_page_url = DOWNLOAD_PAGE_URL.replace("[[EDITION]]", self.edition)

        file_path = folder_path / FILE_NAME
        super().__init__(file_path)

    def _capitalize_edition(self) -> str:
        return "-".join([s.capitalize() for s in self.edition.split("-")])

    @cache
    def _get_download_link(self) -> str:
        latest_version_str = self._version_to_str(self._get_latest_version())
        url = f"{self.download_page_url}/{latest_version_str}"

        edition_page = requests.get(f"{url}?jsontable").json()["data"]

        if any("product" in item["name"] for item in edition_page):
            url += "/product"
        
        if self.edition != "leap-micro":
            latest_version_str += "-NET"

        return f"{url}/iso/openSUSE-{self._capitalize_edition()}-{latest_version_str}-x86_64{"-Current" if self.edition != "leap-micro" else ""}.iso"

    def check_integrity(self) -> bool:
        sha256_url = f"{self._get_download_link()}.sha256"

        sha256_sums = requests.get(sha256_url).text

        sha256_sum = parse_hash(sha256_sums, [], 0)

        return sha256_hash_check(
            self._get_complete_normalized_file_path(absolute=True),
            sha256_sum,
        )

    @cache
    def _get_latest_version(self) -> list[str]:
        r = requests.get(f"{self.download_page_url}?jsontable")

        data = r.json()["data"]

        local_version = self._get_local_version()
        latest = local_version or []

        for i in range(len(data)):
            if "42" in data[i]["name"]:
                continue
            version_number = self._str_to_version(data[i]["name"][:-1])
            if self._compare_version_numbers(latest, version_number):
                sub_r = requests.get(f"{self.download_page_url}/{self._version_to_str(version_number)}?jsontable")
                sub_data = sub_r.json()["data"]
                if not any("iso" in item["name"] or "product" in item["name"] for item in sub_data):
                    continue
                
                latest = version_number

        return latest



from functools import cache
from pathlib import Path

import requests
from bs4 import BeautifulSoup

from modules.exceptions import VersionNotFoundError
from modules.updaters.GenericUpdater import GenericUpdater
from modules.utils import parse_hash, sha256_hash_check

DOMAIN = "https://geo.mirror.pkgbuild.com"
DOWNLOAD_PAGE_URL = f"{DOMAIN}/iso/latest"
FILE_NAME = "archlinux-[[VER]]-x86_64.iso"


class ArchLinux(GenericUpdater):
    """
    A class representing an updater for Arch Linux.

    Attributes:
        download_page (requests.Response): The HTTP response containing the download page HTML.
        soup_download_page (BeautifulSoup): The parsed HTML content of the download page.

    Note:
        This class inherits from the abstract base class GenericUpdater.
    """

    def __init__(self, folder_path: Path) -> None:
        file_path = folder_path / FILE_NAME
        super().__init__(file_path)

        self.download_page = requests.get(DOWNLOAD_PAGE_URL)

        if self.download_page.status_code != 200:
            raise ConnectionError(
                f"Failed to fetch the download page from '{DOWNLOAD_PAGE_URL}'"
            )

        self.soup_download_page = BeautifulSoup(
            self.download_page.content, features="html.parser"
        )

    @cache
    def _get_download_link(self) -> str:
        latest_version_str = self._version_to_str(self._get_latest_version())
        return f"{DOWNLOAD_PAGE_URL}/{FILE_NAME.replace('[[VER]]', latest_version_str)}"

    def check_integrity(self) -> bool:
        sha256_url = "https://geo.mirror.pkgbuild.com/iso/latest/sha256sums.txt"

        sha256_sums = requests.get(sha256_url).text

        sha256_sum = parse_hash(
            sha256_sums,
            [str(self._get_complete_normalized_file_path(absolute=False))],
            0,
        )

        return sha256_hash_check(
            self._get_complete_normalized_file_path(absolute=True),
            sha256_sum,
        )

    @cache
    def _get_latest_version(self) -> list[str]:
        download_a_tags = self.soup_download_page.find_all("a", href=True)
        if not download_a_tags:
            raise VersionNotFoundError("We were not able to parse the download page")
        download_a_tag = next(
            a_tag for a_tag in download_a_tags if "archlinux" in a_tag.get("href")
        )
        if not download_a_tag:
            raise VersionNotFoundError("We were not able to parse the download links")

        return self._str_to_version(download_a_tag.getText().split("-")[1])



import zipfile
from functools import cache
from pathlib import Path

import requests
from bs4 import BeautifulSoup
from bs4.element import Tag

from modules.exceptions import (
    DownloadLinkNotFoundError,
    IntegrityCheckError,
    VersionNotFoundError,
)
from modules.updaters.GenericUpdater import GenericUpdater
from modules.utils import download_file, parse_hash, sha256_hash_check

DOMAIN = "https://www.memtest.org"
DOWNLOAD_PAGE_URL = f"{DOMAIN}"
FILE_NAME = "Memtest86plus-[[VER]].iso"


class MemTest86Plus(GenericUpdater):
    """
    A class representing an updater for MemTest86+.

    Attributes:
        download_page (requests.Response): The HTTP response containing the download page HTML.
        soup_download_page (BeautifulSoup): The parsed HTML content of the download page.
        soup_download_card (Tag): The specific HTML Tag containing the download information card.

    Note:
        This class inherits from the abstract base class GenericUpdater.
    """

    def __init__(self, folder_path: Path) -> None:
        """
        Initialize the MemTest86Plus updater.

        Args:
            folder_path (str): The path to the folder where the MemTest86+ ISO file is stored.

        Raises:
            ConnectionError: If the download page could not be fetched successfully.
            DownloadLinkNotFoundError: If the card containing download information is not found.
        """
        file_path = folder_path / FILE_NAME
        super().__init__(file_path)

        self.download_page = requests.get(DOWNLOAD_PAGE_URL)

        if self.download_page.status_code != 200:
            raise ConnectionError(
                f"Failed to fetch the download page from '{DOWNLOAD_PAGE_URL}'"
            )

        self.soup_download_page = BeautifulSoup(
            self.download_page.content, features="html.parser"
        )
        self.soup_download_card: Tag = self.soup_download_page.find(
            "div", attrs={"class": "col-xxl-4"}
        )  # type: ignore

        if not self.soup_download_card:
            raise DownloadLinkNotFoundError(
                "Could not find the card containing download information"
            )

    @cache
    def _get_download_link(self) -> str:
        download_element: Tag | None = self.soup_download_card.find("a", string="Linux ISO (64 bits)")  # type: ignore
        if not download_element:
            raise DownloadLinkNotFoundError("Could not find the download link")
        return f"{DOWNLOAD_PAGE_URL}{download_element.get('href')}"

    def check_integrity(self) -> bool:
        """
        Check the integrity of the downloaded file by verifying its SHA-256 hash against the one provided on the website.

        Returns:
            bool: True if the integrity check passes, False otherwise.
        """
        version_str = self._version_to_str(self._get_latest_version())
        sha_256_url = f"{DOWNLOAD_PAGE_URL}/download/v{version_str}/sha256sum.txt"
        sha_256_checksums_str: str = requests.get(sha_256_url).text
        sha_256_checksum: str = parse_hash(sha_256_checksums_str, ["64.iso"], 0)

        return sha256_hash_check(
            self._get_complete_normalized_file_path(absolute=True).with_suffix(".zip"),
            sha_256_checksum,
        )

    def install_latest_version(self) -> None:
        """
        Download and install the latest version of the software.

        Raises:
            ValueError: If no download link is available for the latest version.
            IntegrityCheckError: If the integrity check of the downloaded file fails.
        """
        download_link = self._get_download_link()

        new_file = self._get_complete_normalized_file_path(absolute=True)

        archive_path = new_file.with_suffix(".zip")

        download_file(download_link, archive_path)

        local_file = self._get_local_file()

        if not self.check_integrity():
            archive_path.unlink()
            raise IntegrityCheckError("Integrity check failed")

        with zipfile.ZipFile(archive_path) as z:
            file_list = z.namelist()
            iso = next(file for file in file_list if file.endswith(".iso"))
            extracted_file = Path(z.extract(iso, path=new_file.parent))

        if local_file:
            local_file.unlink()
        archive_path.unlink()

        try:
            extracted_file.rename(new_file)
        except FileExistsError:
            # On Windows, files are not overwritten by default, so we need to remove the old file first
            new_file.unlink()
            extracted_file.rename(new_file)

    @cache
    def _get_latest_version(self) -> list[str]:
        card_title: Tag | None = self.soup_download_card.find(
            "span", attrs={"class": "text-primary fs-2"}
        )  # type: ignore

        if not card_title:
            raise VersionNotFoundError("Could not find the latest version")

        return self._str_to_version(
            card_title.getText().split("v")[-1]  # Parse from Memtest86+ v 0.0
        )



from functools import cache
from pathlib import Path

import requests
from bs4 import BeautifulSoup, Tag

from modules.exceptions import VersionNotFoundError
from modules.updaters.GenericUpdater import GenericUpdater
from modules.utils import sha256_hash_check, windows_consumer_download

DOMAIN = "https://www.microsoft.com"
DOWNLOAD_PAGE_URL = f"{DOMAIN}/en-us/software-download/windows11"
FILE_NAME = "Win11_[[VER]]_EnglishInternational_x64v2.iso"


class Windows11(GenericUpdater):
    """
    A class representing an updater for Windows 11.

    Attributes:
        download_page (requests.Response): The HTTP response containing the download page HTML.
        soup_download_page (BeautifulSoup): The parsed HTML content of the download page.

    Note:
        This class inherits from the abstract base class GenericUpdater.
    """

    def __init__(self, folder_path: Path, lang: str) -> None:
        self.valid_langs = [
            "Arabic",
            "Brazilian Portuguese",
            "Bulgarian",
            "Chinese",
            "Chinese",
            "Croatian",
            "Czech",
            "Danish",
            "Dutch",
            "English",
            "English International",
            "Estonian",
            "Finnish",
            "French",
            "French Canadian",
            "German",
            "Greek",
            "Hebrew",
            "Hungarian",
            "Italian",
            "Japanese",
            "Korean",
            "Latvian",
            "Lithuanian",
            "Norwegian",
            "Polish",
            "Portuguese",
            "Romanian",
            "Russian",
            "Serbian Latin",
            "Slovak",
            "Slovenian",
            "Spanish",
            "Spanish (Mexico)",
            "Swedish",
            "Thai",
            "Turkish",
            "Ukrainian",
        ]
        self.lang = lang
        file_path = folder_path / FILE_NAME
        super().__init__(file_path)
        # Make the parameter case insensitive, and find back the correct case using valid_editions
        self.lang = next(
            valid_lang
            for valid_lang in self.valid_langs
            if valid_lang.lower() == self.lang.lower()
        )
        self.version_splitter = "H"
        self.headers = {
            "User-Agent": "Mozilla/5.0 (X11; Linux x86_64; rv:100.0) Gecko/20100101 Firefox/100.0",
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,*/*;q=0.8",
            "referer": "folfy.blue",
        }

        self.download_page = requests.get(DOWNLOAD_PAGE_URL, headers=self.headers)

        if self.download_page.status_code != 200:
            raise ConnectionError(
                f"Failed to fetch the download page from '{DOWNLOAD_PAGE_URL}'"
            )

        self.soup_download_page = BeautifulSoup(
            self.download_page.content, features="html.parser"
        )

        self.hash: str | None = None

    @cache
    def _get_download_link(self) -> str:
        download_link, self.hash = windows_consumer_download(
            windows_version="11", lang=self.lang
        )
        return download_link

    def check_integrity(self) -> bool:
        if not self.hash:
            return False

        return sha256_hash_check(
            self._get_complete_normalized_file_path(absolute=True),
            self.hash,
        )

    @cache
    def _get_latest_version(self) -> list[str]:
        header: Tag | None = self.soup_download_page.find("div", class_="row").find("div").find("p", string=lambda text: "Version" in text)  # type: ignore
        if not header:
            raise VersionNotFoundError(
                "Could not find header containing version information"
            )

        return [
            version_number.strip()
            for version_number in header.getText()
            .split("Version")[1]
            .replace(")", "")
            .split("H")
        ]



import re
from functools import cache
from pathlib import Path

import requests

from modules.exceptions import VersionNotFoundError
from modules.updaters.GenericUpdater import GenericUpdater
from modules.utils import (
    md5_hash_check,
    parse_hash,
    sha256_hash_check,
    sha512_hash_check,
)

DOMAIN = "https://gitlab.manjaro.org"
DOWNLOAD_PAGE_URL = f"{DOMAIN}/web/iso-info/-/raw/master/file-info.json"
FILE_NAME = "manjaro-[[EDITION]]-[[VER]]-linux.iso"


class Manjaro(GenericUpdater):
    """
    A class representing an updater for Manjaro.

    Attributes:
        valid_editions (list[str]): List of valid editions to use
        edition (str): Edition to download
        file_info_json (dict[Any, Any]): JSON file containing file information for each edition

    Note:
        This class inherits from the abstract base class GenericUpdater.
    """

    def __init__(self, folder_path: Path, edition: str) -> None:
        self.valid_editions = [
            "plasma",
            "xfce",
            "gnome",
            "cinnamon",
            "i3",
        ]
        self.edition = edition.lower()
        file_path = folder_path / FILE_NAME
        super().__init__(file_path)

        self.file_info_json = requests.get(DOWNLOAD_PAGE_URL).json()
        self.file_info_json["releases"] = (
            self.file_info_json["official"] | self.file_info_json["community"]
        )

    @cache
    def _get_download_link(self) -> str:
        return self.file_info_json["releases"][self.edition]["image"]

    def check_integrity(self) -> bool:
        checksum_url = self.file_info_json["releases"][self.edition]["checksum"]

        checksums = requests.get(checksum_url).text

        checksum = parse_hash(checksums, [], 0)

        if checksum_url.endswith(".sha512"):
            return sha512_hash_check(
                self._get_complete_normalized_file_path(absolute=True),
                checksum,
            )
        elif checksum_url.endswith(".sha256"):
            return sha256_hash_check(
                self._get_complete_normalized_file_path(absolute=True),
                checksum,
            )
        elif checksum_url.endswith(".md5"):
            return md5_hash_check(
                self._get_complete_normalized_file_path(absolute=True),
                checksum,
            )
        else:
            raise ValueError("Unknown checksum type")

    @cache
    def _get_latest_version(self) -> list[str]:
        download_link = self._get_download_link()

        latest_version_regex = re.search(
            r"manjaro-\w+-(.+?)-",
            download_link,
        )

        if latest_version_regex:
            return self._str_to_version(latest_version_regex.group(1))

        raise VersionNotFoundError("Could not find the latest available version")



class DownloadLinkNotFoundError(Exception):
    """
    Custom exception to represent an error when the download link for a file is not found.
    """

    pass


class VersionNotFoundError(Exception):
    """
    Custom exception to represent an error when a specific version is not found.
    This exception can be raised when a program is unable to find a specific version of a resource.

    Example:
        try:
            version = GenericUpdater._get_latest_version()
        except VersionNotFoundError as e:
            print(f"Error: {e}")
    """

    pass


class IntegrityCheckError(Exception):
    """
    Custom exception to represent an error when an integrity check fails.
    This exception can be raised when a program performs an integrity check (e.g., hash verification) on a file,
    and the check fails, indicating that the file's content may have been altered or corrupted.

    Example:
        if not md5_hash_check(file, expected_md5_hash):
            raise IntegrityCheckError("MD5 hash verification failed.")
    """

    pass



import zipfile
from functools import cache
from pathlib import Path

import requests
from bs4 import BeautifulSoup
from bs4.element import Tag

from modules.exceptions import (
    DownloadLinkNotFoundError,
    IntegrityCheckError,
    VersionNotFoundError,
)
from modules.updaters.GenericUpdater import GenericUpdater
from modules.utils import download_file, parse_hash, sha256_hash_check

DOMAIN = "https://www.supergrubdisk.org"
DOWNLOAD_PAGE_URL = f"{DOMAIN}/category/download/supergrub2diskdownload/"
FILE_NAME = "SuperGrub2-[[VER]].img"


class SuperGrub2(GenericUpdater):
    """
    A class representing an updater for SuperGrub2.

    Attributes:
        download_page (requests.Response): The HTTP response containing the download page HTML.
        soup_download_page (BeautifulSoup): The parsed HTML content of the download page.

    Note:
        This class inherits from the abstract base class GenericUpdater.
    """

    def __init__(self, folder_path: Path) -> None:
        file_path = folder_path / FILE_NAME
        super().__init__(file_path)

        self.download_page = requests.get(DOWNLOAD_PAGE_URL)

        if self.download_page.status_code != 200:
            raise ConnectionError(
                f"Failed to fetch the download page from '{DOWNLOAD_PAGE_URL}'"
            )

        self.soup_download_page = BeautifulSoup(
            self.download_page.content, features="html.parser"
        )

        self.soup_latest_download_article: Tag = self.soup_download_page.find("article")  # type: ignore
        if not self.soup_latest_download_article:
            raise DownloadLinkNotFoundError(
                "Could not find the article containing download information"
            )

    @cache
    def _get_download_link(self) -> str:
        download_tag = self._find_in_table("Download supergrub2")

        if not download_tag:
            raise DownloadLinkNotFoundError(
                "We were not able to find the link to the SourceForge in the table"
            )

        href_attributes = download_tag.find_all(href=True)

        if not href_attributes:
            raise DownloadLinkNotFoundError("No download link found in the `Tag`")

        sourceforge_url = href_attributes[0].get("href")
        return sourceforge_url

    def check_integrity(self, archive_to_check: Path) -> bool:
        sha256_sums_tag = self.soup_latest_download_article.find_all("pre")
        if not sha256_sums_tag:
            raise IntegrityCheckError("Couldn't find the SHA256 sum")
        sha256_sums_tag = sha256_sums_tag[-1]
        sha256_checksums_str = sha256_sums_tag.getText()
        sha_256_checksum: str = parse_hash(
            sha256_checksums_str,
            [
                f"supergrub2-{self._get_latest_version()[0]}",
                ".img.zip",
            ],
            0,
        )

        return sha256_hash_check(archive_to_check, sha_256_checksum)

    def install_latest_version(self) -> None:
        download_link: str = self._get_download_link()

        new_file = self._get_complete_normalized_file_path(absolute=True)

        archive_path = new_file.with_suffix(".zip")

        download_file(download_link, archive_path)

        local_file = self._get_local_file()

        if not self.check_integrity(archive_path):
            archive_path.unlink()
            raise IntegrityCheckError("Integrity check failed")

        with zipfile.ZipFile(archive_path) as z:
            file_list = z.namelist()
            iso = next(file for file in file_list if file.endswith(".img"))
            extracted_file = Path(z.extract(iso, path=new_file.parent))

        if local_file:
            local_file.unlink()
        archive_path.unlink()

        try:
            extracted_file.rename(new_file)
        except FileExistsError:
            # On Windows, files are not overwritten by default, so we need to remove the old file first
            new_file.unlink()
            extracted_file.rename(new_file)

    @cache
    def _get_latest_version(self) -> list[str]:
        download_table: Tag | None = self.soup_latest_download_article.find("table", attrs={"cellpadding": "5px"})  # type: ignore
        if not download_table:
            raise VersionNotFoundError(
                "We were not able to find the table of download which contains the version number"
            )

        download_table_header: Tag | None = download_table.find("h2")  # type: ignore
        if not download_table_header:
            raise VersionNotFoundError(
                "We were not able to find the header containing the version number"
            )

        header: str = download_table_header.getText().lower()
        return self._str_to_version(
            header.replace("super grub2 disk", "")
            .strip()  # Parse from "Super Grub2 Disk 2.06s2-beta1"
            .replace("s", self.version_splitter)
            .replace("-beta", self.version_splitter)
        )

    def _find_in_table(self, row_name_contains: str) -> Tag | None:
        """
        Find the HTML Tag containing specific information in the download page table.

        Args:
            row_name_contains (str): A string that identifies the row in the table.

        Returns:
            Tag | None: The HTML Tag containing the desired information, or None .
        """
        download_table: Tag | None = self.soup_latest_download_article.find("table", attrs={"cellpadding": "5px"})  # type: ignore

        if not download_table:
            raise LookupError("Could not find the table with download information")

        for tr in download_table.find_all("tr"):
            for td in tr.find_all("td"):
                if row_name_contains in td.getText():
                    return td
        return None



import re
from functools import cache
from pathlib import Path

import requests
from bs4 import BeautifulSoup
from bs4.element import Tag

from modules.exceptions import DownloadLinkNotFoundError, VersionNotFoundError
from modules.updaters.GenericUpdater import GenericUpdater
from modules.utils import parse_hash, sha256_hash_check

DOMAIN = "https://www.system-rescue.org"
DOWNLOAD_PAGE_URL = f"{DOMAIN}/Download"
FILE_NAME = "systemrescue-[[VER]]-amd64.iso"


class SystemRescue(GenericUpdater):
    """
    A class representing an updater for SystemRescue.

    Attributes:
        download_page (requests.Response): The HTTP response containing the download page HTML.
        soup_download_page (BeautifulSoup): The parsed HTML content of the download page.

    Note:
        This class inherits from the abstract base class GenericUpdater.
    """

    def __init__(self, folder_path: Path) -> None:
        """
        Initialize a SystemRescue updater object.

        Args:
            folder_path (Path): The path to the folder where the SystemRescue file is stored.

        Raises:
            ConnectionError: If the download page cannot be fetched successfully.
        """
        file_path = folder_path / FILE_NAME
        super().__init__(file_path)

        self.download_page = requests.get(DOWNLOAD_PAGE_URL)

        if self.download_page.status_code != 200:
            raise ConnectionError(
                f"Failed to fetch the download page from '{DOWNLOAD_PAGE_URL}'"
            )

        self.soup_download_page = BeautifulSoup(
            self.download_page.content, features="html.parser"
        )

    @cache
    def _get_download_link(self) -> str:
        download_tag: Tag | None = self._find_in_table("Fastly")

        if not download_tag:
            raise DownloadLinkNotFoundError(
                "Failed to find the `Tag` containing the download link"
            )

        href_attributes = download_tag.find_all(href=True)
        if not href_attributes:
            raise DownloadLinkNotFoundError("No download link found in the `Tag`")

        return href_attributes[0].get("href")

    def check_integrity(self) -> bool:
        version_str = self._version_to_str(self._get_latest_version())
        sha256_download_link = f"{DOMAIN}/releases/{version_str}/systemrescue-{version_str}-amd64.iso.sha256"

        r = requests.get(sha256_download_link)
        sha256_checksum = parse_hash(
            r.text,
            [str(self._get_normalized_file_path(False, self._get_latest_version()))],
            0,
        )

        return sha256_hash_check(
            self._get_complete_normalized_file_path(absolute=True),
            sha256_checksum,
        )

    @cache
    def _get_latest_version(self) -> list[str]:
        download_link = self._get_download_link()

        latest_version_regex = re.search(
            r"releases\/(.+)\/",  # Parse from https://fastly-cdn.system-rescue.org/releases/10.01/systemrescue-10.01-amd64.iso
            download_link,
        )

        if latest_version_regex:
            return self._str_to_version(latest_version_regex.group(1))

        raise VersionNotFoundError("Could not find the latest available version")

    def _find_in_table(self, row_name_contains: str) -> Tag | None:
        """
        Find the HTML Tag containing specific information in the download page table.

        Args:
            row_name_contains (str): A string that identifies the row in the table.

        Returns:
            Tag | None: The HTML Tag containing the desired information, or None if not found.

        Raises:
            LookupError: If the table or the specified row_name_contains is not found in the download page.
        """
        s: Tag | None = self.soup_download_page.find("div", attrs={"id": "colcenter"})  # type: ignore

        if not s:
            raise LookupError(
                "Could not find the div containing the table with download information"
            )

        s = s.find("table")  # type: ignore

        if not s:
            raise LookupError(
                "Could not find the table containing download information"
            )

        for tr in s.find_all("tr"):
            for td in tr.find_all("td"):
                if row_name_contains in td.getText():
                    return td

        raise LookupError(f"Failed to find '{row_name_contains}' in the table")



from functools import cache
from pathlib import Path

import requests

from modules.updaters.GenericUpdater import GenericUpdater
from modules.updaters.util_update_checker import (
    github_get_latest_version,
    parse_github_release,
)
from modules.utils import parse_hash, sha256_hash_check

FILE_NAME = "rescuezilla-[[VER]]-64bit.[[EDITION]].iso"


class Rescuezilla(GenericUpdater):
    """
    A class representing an updater for Rescuezilla.

    Attributes:
        valid_editions (list[str]): List of valid editions to use
        edition (str): Edition to download
        release_info (dict): Github release information

    Note:
        This class inherits from the abstract base class GenericUpdater.
    """

    def __init__(self, folder_path: Path, edition: str) -> None:
        self.valid_editions = ["focal", "jammy", "lunar", "mantic", "noble"]
        self.edition = edition.lower()

        file_path = folder_path / FILE_NAME
        super().__init__(file_path)

        release = github_get_latest_version("rescuezilla", "rescuezilla")

        self.release_info = parse_github_release(release)

    @cache
    def _get_download_link(self) -> str:
        return self.release_info["files"][
            str(self._get_complete_normalized_file_path(absolute=False))
        ]

    def check_integrity(self) -> bool:
        sha256_url = self.release_info["files"]["SHA256SUM"]

        sha256_sums = requests.get(sha256_url).text

        sha256_sum = parse_hash(
            sha256_sums,
            [str(self._get_complete_normalized_file_path(absolute=False))],
            0,
        )

        return sha256_hash_check(
            self._get_complete_normalized_file_path(absolute=True),
            sha256_sum,
        )

    @cache
    def _get_latest_version(self) -> list[str]:
        return self._str_to_version(self.release_info["tag"])



"""Based off of pypa/sampleproject
https://raw.githubusercontent.com/pypa/sampleproject/db5806e0a3204034c51b1c00dde7d5eb3fa2532e/setup.py
"""

import pathlib

# Always prefer setuptools over distutils
from setuptools import find_packages, setup

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
    name="sisou",  # Required
    version="1.1.9",  # Required
    description="A powerful tool to conveniently update all of your ISO files!",  # Optional
    long_description=long_description,  # Optional
    long_description_content_type="text/markdown",  # Optional
    url="https://github.com/JoshuaVandaele/SuperISOUpdater",  # Optional
    author="Joshua Vandaele",  # Optional
    author_email="joshua@vandaele.software",  # Optional
    classifiers=[  # Optional
        # https://pypi.org/classifiers/
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: End Users/Desktop",
        "Topic :: Internet",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate you support Python 3. These classifiers are *not*
        # checked by 'pip install'. See instead 'python_requires' below.
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3 :: Only",
    ],
    keywords="ventoy, updater, os, iso, updater, sisou, cli",  # Optional
    packages=find_packages(),  # Required
    py_modules=["sisou"],  # Required
    include_package_data=True,
    package_data={"": ["config.toml.default"]},
    python_requires=">=3.10, <4",
    install_requires=[
        "beautifulsoup4==4.12.2",
        "requests==2.31.0",
        "tqdm==4.65.0",
        "PGPy==0.6.0",
    ],  # Optional
    # extras_require={
    #     "dev": [""],
    #     "test": [""],
    # },
    entry_points={  # Optional
        "console_scripts": [
            "sisou = sisou:main",
        ],
    },
    project_urls={  # Optional
        "Bug Reports": "https://github.com/JoshuaVandaele/SuperISOUpdater/issues",
        "Source": "https://github.com/JoshuaVandaele/SuperISOUpdater/",
    },
    zip_safe=False,
)



from functools import cache
from pathlib import Path

import requests
from bs4 import BeautifulSoup

from modules.updaters.GenericUpdater import GenericUpdater
from modules.utils import sha256_hash_check

DOMAIN = "https://www.truenas.com"
DOWNLOAD_PAGE_URL = f"{DOMAIN}/download-truenas-[[EDITION]]"
FILE_NAME = "TrueNAS-[[EDITION]]-[[VER]].iso"


class TrueNAS(GenericUpdater):
    """
    A class representing an updater for TrueNAS.

    Attributes:
        valid_editions (list[str]): List of valid editions to use
        edition (str): Edition to download
        download_page (requests.Response): The HTTP response containing the download page HTML.
        soup_download_page (BeautifulSoup): The parsed HTML content of the download page.

    Note:
        This class inherits from the abstract base class GenericUpdater.
    """

    def __init__(self, folder_path: Path, edition: str) -> None:
        self.valid_editions = ["core", "scale"]
        self.edition = edition.lower()

        file_path = folder_path / FILE_NAME
        super().__init__(file_path)

        self.download_page_url = DOWNLOAD_PAGE_URL.replace("[[EDITION]]", self.edition)

        self.download_page = requests.get(self.download_page_url)

        if self.download_page.status_code != 200:
            raise ConnectionError(
                f"Failed to fetch the download page from '{DOWNLOAD_PAGE_URL}'"
            )

        self.soup_download_page = BeautifulSoup(
            self.download_page.content, features="html.parser"
        )

    @cache
    def _get_download_link(self) -> str:
        a_tag = self.soup_download_page.find("a", attrs={"id": "downloadTrueNAS"})

        if not a_tag:
            raise LookupError("Could not find HTML tag containing download URL")

        return a_tag["href"]  # type: ignore

    def check_integrity(self) -> bool:
        sha256_url = f"{self._get_download_link()}.sha256"

        # Only 1 sum in file
        sha256_sums = requests.get(sha256_url).text.split()

        # for some reason TrueNAS has two different formats for their sha256 file
        if sha256_sums[0] == "SHA256":
            sha256_sum = sha256_sums[-1]
        else:
            sha256_sum = sha256_sums[0]

        return sha256_hash_check(
            self._get_complete_normalized_file_path(absolute=True),
            sha256_sum,
        )

    @cache
    def _get_latest_version(self) -> list[str]:
        download_link = self._get_download_link()
        version = download_link.split("-")[-1]

        return self._str_to_version(version.replace(".iso", ""))



import argparse
import logging
from abc import ABCMeta
from functools import cache
from pathlib import Path
from typing import Type

import modules.updaters
from modules.updaters import GenericUpdater
from modules.utils import parse_config


@cache
def get_available_updaters() -> list[Type[GenericUpdater]]:
    """Get a list of available updaters.

    Returns:
        list[Type[GenericUpdater]]: A list of available updater classes.
    """
    return [
        getattr(modules.updaters, updater)
        for updater in dir(modules.updaters)
        if updater != "GenericUpdater"
        and isinstance(getattr(modules.updaters, updater), ABCMeta)
        and issubclass(getattr(modules.updaters, updater), GenericUpdater)
    ]


def setup_logging(log_level: str, log_file: Path | None):
    """Set up logging configurations.

    Args:
        log_level (str): The log level. Valid choices: "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL".
        log_file (Path | None): The path to the log file. If None, log to console.

    Raises:
        ValueError: If the log_level is invalid.
    """
    numeric_log_level = getattr(logging, log_level, None)

    logging.basicConfig(
        level=numeric_log_level,
        format="%(asctime)s - %(levelname)s - %(message)s",
        filename=log_file,
    )

    logging.debug("Logging started")


def run_updater(updater: GenericUpdater):
    """Run a single updater.

    Args:
        updater (GenericUpdater): The updater instance to run.
    """
    installer_for = f"{updater.__class__.__name__}{' '+updater.edition if updater.has_edition() else ''}"  # type: ignore

    logging.info(f"[{installer_for}] Checking for updates...")

    try:
        if updater.check_for_updates():
            logging.info(
                f"[{installer_for}] Updates available. Downloading and installing the latest version..."
            )
            updater.install_latest_version()
            logging.info(f"[{installer_for}] Update completed successfully!")
        else:
            logging.info(f"[{installer_for}] No updates available.")
    except:
        logging.exception(
            f"[{installer_for}] An error occurred while updating. See traceback below."
        )


def run_updaters(
    install_path: Path, config: dict, updater_list: list[Type[GenericUpdater]]
):
    """Run updaters based on the provided configuration.

    Args:
        install_path (Path): The installation path.
        config (dict): The configuration dictionary.
        updater_list (list[Type[GenericUpdater]]): A list of available updater classes.
    """
    for key, value in config.items():
        # If the key's name is the name of an updater, run said updater using the values as argument, otherwise assume it's a folder's name
        if key in [updater.__name__ for updater in updater_list]:
            updater_class = next(
                updater for updater in updater_list if updater.__name__ == key
            )

            updaters: list[GenericUpdater] = []

            params: list[dict] = [{}]

            editions = value.get("editions", [])
            langs = value.get("langs", [])

            if editions and langs:
                params = [
                    {"edition": edition, "lang": lang}
                    for edition in editions
                    for lang in langs
                ]
            elif editions:
                params = [{"edition": edition} for edition in editions]
            elif langs:
                params = [{"lang": lang} for lang in langs]

            for param in params:
                try:
                    updaters.append(updater_class(install_path, **param))
                except Exception:
                    installer_for = f"{key} {param}"
                    logging.exception(
                        f"[{installer_for}] An error occurred while trying to add the installer. See traceback below."
                    )
            # Run updater(s)
            for updater in updaters:
                run_updater(updater)

        else:
            run_updaters(install_path / key, value, updater_list)


def main():
    """Main function to run the update process."""
    parser = argparse.ArgumentParser(description="Process a file and set log level")

    # Add the positional argument for the file path
    parser.add_argument("ventoy_path", help="Path to the Ventoy drive")

    # Add the optional argument for log level
    parser.add_argument(
        "-l",
        "--log-level",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default="INFO",
        help="Set the log level (default: INFO)",
    )

    # Add the optional argument for log file
    parser.add_argument(
        "-f", "--log-file", help="Path to the log file (default: log to console)"
    )

    # Add the optional argument for config file
    parser.add_argument(
        "-c", "--config-file", help="Path to the config file (default: config.toml)"
    )

    args = parser.parse_args()

    log_file = Path(args.log_file) if args.log_file else None
    setup_logging(args.log_level, log_file)

    ventoy_path = Path(args.ventoy_path).resolve()

    config_file = Path(args.config_file) if args.config_file else None
    if not config_file:
        logging.info(
            "No config file specified. Trying to find config.toml in the current directory..."
        )
        config_file = Path() / "config.toml"

        if not config_file.is_file():
            logging.info(
                "No config file specified. Trying to find config.toml in the ventoy drive..."
            )
            config_file = ventoy_path / "config.toml"

            if not config_file.is_file():
                logging.info(
                    "No config.toml found in the ventoy drive. Generating one from config.toml.default..."
                )
                with open(
                    Path(__file__).parent / "config" / "config.toml.default"
                ) as default_config_file:
                    config_file.parent.mkdir(parents=True, exist_ok=True)
                    with open(config_file, "w") as new_config_file:
                        new_config_file.write(default_config_file.read())
                logging.info(
                    "Generated config.toml in the ventoy drive. Please edit it to your liking and run sisou again."
                )
                return

    config = parse_config(config_file)
    if not config:
        raise ValueError("Configuration file could not be parsed or is empty")

    available_updaters: list[Type[GenericUpdater]] = get_available_updaters()

    run_updaters(ventoy_path, config, available_updaters)

    logging.debug("Finished execution")


if __name__ == "__main__":
    main()



import glob
import logging
import re
import zipfile
from functools import cache
from pathlib import Path

import requests
from bs4 import BeautifulSoup

from modules.exceptions import IntegrityCheckError, VersionNotFoundError
from modules.updaters.GenericUpdater import GenericUpdater
from modules.utils import download_file, parse_hash, sha256_hash_check

DOMAIN = "https://www.ibiblio.org"
DOWNLOAD_PAGE_URL = f"{DOMAIN}/pub/micro/pc-stuff/freedos/files/distributions/"
FILE_NAME = "FreeDOS-[[VER]]-[[EDITION]].[[EXT]]"


class FreeDOS(GenericUpdater):
    """
    A class representing an updater for FreeDOS.

    Attributes:
        download_page (requests.Response): The HTTP response containing the download page HTML.
        soup_download_page (BeautifulSoup): The parsed HTML content of the download page.

    Note:
        This class inherits from the abstract base class GenericUpdater.
    """

    def __init__(self, folder_path: Path, edition: str) -> None:
        self.valid_editions = [
            "BonusCD",
            "FloppyEdition",
            "FullUSB",
            "LegacyCD",
            "LiteUSB",
            "LiveCD",
        ]

        self.edition = edition
        file_path = folder_path / FILE_NAME
        super().__init__(file_path)

        # Make the parameter case insensitive, and find back the correct case using valid_editions
        self.edition = next(
            valid_ed
            for valid_ed in self.valid_editions
            if valid_ed.lower() == self.edition.lower()
        )

        self.download_page = requests.get(DOWNLOAD_PAGE_URL)

        if self.download_page.status_code != 200:
            raise ConnectionError(
                f"Failed to fetch the download page from '{DOWNLOAD_PAGE_URL}'"
            )

        self.soup_download_page = BeautifulSoup(
            self.download_page.content, features="html.parser"
        )

    @cache
    def _get_download_link(self) -> str:
        latest_version = self._get_latest_version()
        latest_version_str = self._version_to_str(latest_version)
        return f"{DOWNLOAD_PAGE_URL}/{latest_version_str}/official/FD{''.join(latest_version)}-{self.edition}.zip"

    def check_integrity(self) -> bool:
        checksums_url = "https://www.ibiblio.org/pub/micro/pc-stuff/freedos/files/distributions/1.3/official/verify.txt"

        checksums = requests.get(checksums_url).text

        try:
            sha256_sums = next(
                sums for sums in checksums.split("\n\n") if "sha256" in sums
            )
        except StopIteration as e:
            raise IntegrityCheckError(
                "Could not find the sha256 hash in the hash list file"
            ) from e

        sha256_sum = parse_hash(sha256_sums, [self.edition], 0)

        return sha256_hash_check(
            self._get_normalized_file_path(
                True, self._get_latest_version(), self.edition
            ).with_suffix(".zip"),
            sha256_sum,
        )

    def install_latest_version(self) -> None:
        """
        Download and install the latest version of the software.

        Raises:
            IntegrityCheckError: If the integrity check of the downloaded file fails.
        """
        download_link = self._get_download_link()

        new_file = self._get_complete_normalized_file_path(absolute=True)
        archive_path = new_file.with_suffix(".zip")

        local_file = self._get_local_file()

        download_file(download_link, archive_path)

        try:
            integrity_check = self.check_integrity()
        except IntegrityCheckError as e:
            raise e
        except Exception as e:
            raise IntegrityCheckError(
                "Integrity check failed: An error occurred"
            ) from e

        if not integrity_check:
            archive_path.unlink()
            raise IntegrityCheckError("Integrity check failed: Hashes do not match")

        with zipfile.ZipFile(archive_path) as z:
            file_list = z.namelist()
            try:
                file_ext = "ISO"
                to_extract = next(
                    file for file in file_list if file.upper().endswith(file_ext)
                )
            except StopIteration:
                file_ext = "IMG"
                to_extract = next(
                    file for file in file_list if file.upper().endswith(file_ext)
                )

            extracted_file = Path(z.extract(to_extract, path=new_file.parent))
        try:
            extracted_file.rename(new_file.with_suffix(file_ext))
        except FileExistsError:
            # On Windows, files are not overwritten by default, so we need to remove the old file first
            new_file.unlink()
            extracted_file.rename(new_file.with_suffix(file_ext))

        archive_path.unlink()
        if local_file:
            os.remove(local_file)  # type: ignore

    def _get_local_file(self) -> Path | None:
        file_path = self._get_normalized_file_path(
            absolute=True,
            version=None,
            edition=self.edition if self.has_edition() else None,  # type: ignore
            lang=self.lang if self.has_lang() else None,  # type: ignore
        )

        local_files = glob.glob(
            str(file_path.with_suffix(".*")).replace("[[VER]]", "*")
        )

        if local_files:
            return Path(local_files[0])
        logging.debug(
            f"[FreeDOS._get_local_file] No local file found for {self.__class__.__name__}"
        )
        return None

    @cache
    def _get_latest_version(self) -> list[str]:
        download_a_tags = self.soup_download_page.find_all("a", href=True)
        if not download_a_tags:
            raise VersionNotFoundError("We were not able to parse the download page")

        latest_version = self._get_local_version()
        version_regex = re.compile(r"^(([0-9]+)(\.?))+$")
        for a_tag in download_a_tags:
            href = a_tag.get("href")
            version: str = href[:-1]
            if version_regex.fullmatch(version):
                compared_version = self._str_to_version(version)
                if latest_version:
                    if self._compare_version_numbers(latest_version, compared_version):
                        latest_version = compared_version
                else:
                    latest_version = compared_version

        if not latest_version:
            raise VersionNotFoundError("Could not find a valid version")

        return latest_version
